#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para obtener Código de Trazabilidad de Granos
del web service WSCTG versión 3.0 de AFIP (RG3593/14)
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2010-2014 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.13b"

LICENCIA = """
wsctgv2.py: Interfaz para generar Código de Trazabilidad de Granos AFIP v1.1
Copyright (C) 2014-2015 Mariano Reingart reingart@gmail.com
http://www.sistemasagiles.com.ar/trac/wiki/CodigoTrazabilidadGranos

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

AYUDA="""
Opciones:
  --ayuda: este mensaje

  --debug: modo depuración (detalla y confirma las operaciones)
  --formato: muestra el formato de los archivos de entrada/salida
  --prueba: genera y autoriza una CTG de prueba (no usar en producción!)
  --xml: almacena los requerimientos y respuestas XML (depuración)

  --dummy: consulta estado de servidores
  --solicitar: obtiene el CTG (según archivo de entrada en TXT o CSV)
  --confirmar: confirma el CTG (según archivo de entrada en TXT o CSV)
  --anular: anula el CTG
  --rechazar: permite al destino rechazar el CTG
  --confirmar_arribo: confirma el arribo de un CTG
  --confirmar_definitivo: confirma el arribo definitivo de un CTG
  --regresar_a_origen_rechazado: tomar la acción de "Regresar a Origen"
  --cambiar_destino_destinatario_rechazado: "Cambio de Destino y Destinatario"

  --consultar: consulta las CTG generadas
  --consultar_excel: consulta las CTG generadas (genera un excel)
  --consultar_detalle: obtiene el detalle de una CTG
  --consultar_constancia_pdf: descarga el documento PDF de una CTG
  --pendientes: consulta CTGs otorgados, rechazados, confirmados a resolver
  --consultar_rechazados: obtener CTGs rechazados para darles un nuevo curso
  --consultar_activos_por_patente: consulta de CTGs activos por patente

  --provincias: obtiene el listado de provincias
  --localidades: obtiene el listado de localidades por provincia
  --especies: obtiene el listado de especies
  --cosechas: obtiene el listado de cosechas

Ver wsctg.ini para parámetros de configuración (URL, certificados, etc.)"
"""

import os, sys, time, base64
from php import date
import traceback
from py3simplesoap.client import SoapFault
import utils

# importo funciones compartidas:
from utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I, json, BaseWS, inicializar_y_capturar_excepciones, get_install_dir


# constantes de configuración (homologación):

WSDL = "https://fwshomo.afip.gov.ar/wsctg/services/CTGService_v3.0?wsdl"

DEBUG = False
XML = False
CONFIG_FILE = "wsctg.ini"
HOMO = False

# definición del formato del archivo de intercambio:

ENCABEZADO = [
    # datos enviados
    ('tipo_reg', 1, A), # 0: encabezado
    ('numero_carta_de_porte', 13, N),
    ('codigo_especie', 5, N),
    ('cuit_canjeador', 11, N),
    ('cuit_destino', 11, N),
    ('cuit_destinatario', 11, N),
    ('codigo_localidad_origen', 6, N),
    ('codigo_localidad_destino', 6, N),
    ('codigo_cosecha', 4, N),
    ('peso_neto_carga', 5, N),
    ('cant_horas', 2, N),
    ('patente_vehiculo', 6, A),
    ('cuit_transportista', 11, N),
    ('km_a_recorrer', 4, N),                          # km_recorridos (en consulta WSCTGv2)
    ('establecimiento', 6, N),                        # confirmar arribo
    ('remitente_comercial_como_canjeador', 1, A),     # S/N solicitar CTG inicial (WSCTGv2)
    ('consumo_propio', 1, A),                         # S/N confirmar arribo (WSCTGv2)

    # datos devueltos
    ('numero_ctg', 8, N),
    ('fecha_hora', 19, A),
    ('vigencia_desde', 10, A),
    ('vigencia_hasta', 10, A),
    ('transaccion', 12, N),
    ('tarifa_referencia', 6, I, 2),           # consultar detalle
    ('estado', 20, A),
    ('imprime_constancia', 5, A),
    ('observaciones', 200, A),
    ('errores', 1000, A),
    ('controles', 1000, A),
    ('detalle', 1000, A),                     # consultar detalle (WSCTGv2)

    # nuevos campos agregados:
    ('cuit_chofer', 11, N),

    # nuevos campos agregados WSCTGv3:
    ('cuit_corredor', 12, N),
    ('remitente_comercial_como_productor', 1, A),

    ]



class WSCTGv2(BaseWS):
    "Interfaz para el WebService de Código de Trazabilidad de Granos (Version 3)"
    _public_methods_ = ['Conectar', 'Dummy', 'SetTicketAcceso', 'DebugLog',
                        'SolicitarCTGInicial', 'SolicitarCTGDatoPendiente',
                        'ConfirmarArribo', 'ConfirmarDefinitivo',
                        'AnularCTG', 'RechazarCTG', 'CTGsPendientesResolucion',
                        'ConsultarCTG', 'LeerDatosCTG', 'ConsultarDetalleCTG',
                        'ConsultarCTGExcel', 'ConsultarConstanciaCTGPDF',
                        'ConsultarCTGRechazados',
                        'RegresarAOrigenCTGRechazado',
                        'CambiarDestinoDestinatarioCTGRechazado',
                        'ConsultarCTGActivosPorPatente',
                        'ConsultarProvincias',
                        'ConsultarLocalidadesPorProvincia',
                        'ConsultarEstablecimientos',
                        'ConsultarCosechas',
                        'ConsultarEspecies',
                        'SetParametros', 'SetParametro', 'GetParametro',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        ]
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'Excepcion', 'ErrCode', 'ErrMsg', 'LanzarExcepciones', 'Errores',
        'XmlRequest', 'XmlResponse', 'Version', 'Traceback',
        'NumeroCTG', 'CartaPorte', 'FechaHora', 'CodigoOperacion',
        'CodigoTransaccion', 'Observaciones', 'Controles', 'DatosCTG',
        'VigenciaHasta', 'VigenciaDesde', 'Estado', 'ImprimeConstancia',
        'TarifaReferencia', 'Destino', 'Destinatario', 'Detalle',
        'Patente', 'PesoNeto', 'FechaVencimiento',
        'UsuarioSolicitante', 'UsuarioReal',
        ]
    _reg_progid_ = "WSCTGv2"
    _reg_clsid_ = "{ACDEFB8A-34E1-48CF-94E8-6AF6ADA0717A}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    LanzarExcepciones = False
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')

    def Conectar(self, *args, **kwargs):
        ret = BaseWS.Conectar(self, *args, **kwargs)
        # corregir descripción de servicio WSDL publicado por AFIP
        # kmARecorrer -> kmRecorridos (ConsultarDetalleCTG)
        port = self.client.services['CTGService_v3.0']['ports']['CTGServiceHttpSoap20Endpoint']
        msg = port['operations']['consultarDetalleCTG']['output']['consultarDetalleCTGResponse']
        msg['response']['consultarDetalleCTGDatos']['kmRecorridos'] = int
        return ret

    def inicializar(self):
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.CodError = self.DescError = ''
        self.NumeroCTG = self.CartaPorte = ""
        self.CodigoTransaccion = self.Observaciones = ''
        self.FechaHora = self.CodigoOperacion = ""
        self.VigenciaDesde = self.VigenciaHasta = ""
        self.Controles = []
        self.DatosCTG = self.TarifaReferencia = None
        self.CodigoTransaccion = self.Observaciones = ''
        self.Detalle = self.Destino = self.Destinatario = ''
        self.Patente = self.PesoNeto = self.FechaVencimiento = ''
        self.UsuarioSolicitante = self.UsuarioReal = ''


    def __analizar_errores(self, ret):
        "Comprueba y extrae errores si existen en la respuesta XML"
        if 'arrayErrores' in ret:
            errores = ret['arrayErrores'] or []
            self.Errores = [err['error'] for err in errores]
            self.ErrCode = ' '.join(self.Errores)
            self.ErrMsg = '\n'.join(self.Errores)

    def __analizar_controles(self, ret):
        "Comprueba y extrae controles si existen en la respuesta XML"
        if 'arrayControles' in ret:
            controles = ret['arrayControles']
            self.Controles = ["%(tipo)s: %(descripcion)s" % ctl['control']
                                for ctl in controles]

    @inicializar_y_capturar_excepciones
    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        results = self.client.dummy()['response']
        self.AppServerStatus = str(results['appserver'])
        self.DbServerStatus = str(results['dbserver'])
        self.AuthServerStatus = str(results['authserver'])

    @inicializar_y_capturar_excepciones
    def AnularCTG(self, carta_porte, ctg):
        "Anular el CTG si se creó el mismo por error"
        response = self.client.anularCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosAnularCTG={
                            'cartaPorte': carta_porte,
                            'ctg': ctg, }))['response']
        datos = response.get('datosResponse')
        self.__analizar_errores(response)
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['CTG'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoOperacion = str(datos['codigoOperacion'])

    @inicializar_y_capturar_excepciones
    def RechazarCTG(self, carta_porte, ctg, motivo):
        "El Destino puede rechazar el CTG a través de la siguiente operatoria"
        response = self.client.rechazarCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosRechazarCTG={
                            'cartaPorte': carta_porte,
                            'ctg': ctg, 'motivoRechazo': motivo,
                            }))['response']
        datos = response.get('datosResponse')
        self.__analizar_errores(response)
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['CTG'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoOperacion = str(datos['codigoOperacion'])

    @inicializar_y_capturar_excepciones
    def SolicitarCTGInicial(self, numero_carta_de_porte, codigo_especie,
        cuit_canjeador, cuit_destino, cuit_destinatario, codigo_localidad_origen,
        codigo_localidad_destino, codigo_cosecha, peso_neto_carga,
        cant_horas=None, patente_vehiculo=None, cuit_transportista=None,
        km_a_recorrer=None, remitente_comercial_como_canjeador=None,
        cuit_corredor=None, remitente_comercial_como_productor=None,
        **kwargs):
        "Solicitar CTG Desde el Inicio"
        # ajusto parámetros según validaciones de AFIP:
        if cuit_canjeador and int(cuit_canjeador) == 0:
            cuit_canjeador = None         # nulo
        if not remitente_comercial_como_canjeador:
            remitente_comercial_como_canjeador = None

        ret = self.client.solicitarCTGInicial(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                         datosSolicitarCTGInicial=dict(
                            cartaPorte=numero_carta_de_porte,
                            codigoEspecie=codigo_especie,
                            cuitCanjeador=cuit_canjeador or None,
                            remitenteComercialComoCanjeador=remitente_comercial_como_canjeador,
                            cuitDestino=cuit_destino,
                            cuitDestinatario=cuit_destinatario,
                            codigoLocalidadOrigen=codigo_localidad_origen,
                            codigoLocalidadDestino=codigo_localidad_destino,
                            codigoCosecha=codigo_cosecha,
                            pesoNeto=peso_neto_carga,
                            cuitTransportista=cuit_transportista,
                            cantHoras=cant_horas,
                            patente=patente_vehiculo,
                            kmARecorrer=km_a_recorrer,
                            cuitCorredor=cuit_corredor,
                            remitenteComercialcomoProductor=remitente_comercial_como_productor,
                            )))['response']
        self.__analizar_errores(ret)
        self.Observaciones = ret['observacion']
        datos = ret.get('datosSolicitarCTGResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            datos_ctg = datos.get('datosSolicitarCTG')
            if datos_ctg:
                self.NumeroCTG = str(datos_ctg['ctg'])
                self.FechaHora = str(datos_ctg['fechaEmision'])
                self.VigenciaDesde = str(datos_ctg['fechaVigenciaDesde'])
                self.VigenciaHasta = str(datos_ctg['fechaVigenciaHasta'])
                self.TarifaReferencia = str(datos_ctg.get('tarifaReferencia'))
            self.__analizar_controles(datos)
        return self.NumeroCTG or 0

    @inicializar_y_capturar_excepciones
    def SolicitarCTGDatoPendiente(self, numero_carta_de_porte, cant_horas,
        patente_vehiculo, cuit_transportista):
        "Solicitud que permite completar los datos faltantes de un Pre-CTG "
        "generado anteriormente a través de la operación solicitarCTGInicial"
        ret = self.client.solicitarCTGDatoPendiente(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosSolicitarCTGDatoPendiente=dict(
                            cartaPorte=numero_carta_de_porte,
                            cuitTransportista=cuit_transportista,
                            cantHoras=cant_horas,
                            )))['response']
        self.__analizar_errores(ret)
        self.Observaciones = ret['observacion']
        datos = ret.get('datosSolicitarCTGResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            datos_ctg = datos.get('datosSolicitarCTG')
            if datos_ctg:
                self.NumeroCTG = str(datos_ctg['ctg'])
                self.FechaHora = str(datos_ctg['fechaEmision'])
                self.VigenciaDesde = str(datos_ctg['fechaVigenciaDesde'])
                self.VigenciaHasta = str(datos_ctg['fechaVigenciaHasta'])
                self.TarifaReferencia = str(datos_ctg.get('tarifaReferencia'))
            self.__analizar_controles(datos)
        return self.NumeroCTG

    @inicializar_y_capturar_excepciones
    def ConfirmarArribo(self, numero_carta_de_porte, numero_ctg,
                        cuit_transportista, peso_neto_carga,
                        consumo_propio, establecimiento=None, cuit_chofer=None,
                        **kwargs):
        "Confirma arribo CTG"
        ret = self.client.confirmarArribo(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosConfirmarArribo=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            cuitTransportista=cuit_transportista,
                            cuitChofer=cuit_chofer,
                            cantKilosCartaPorte=peso_neto_carga,
                            consumoPropio=consumo_propio,
                            establecimiento=establecimiento,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos['codigoOperacion'])
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def ConfirmarDefinitivo(self, numero_carta_de_porte, numero_ctg,
                            establecimiento=None, codigo_cosecha=None, peso_neto_carga=None,
                            **kwargs):
        "Confirma arribo definitivo CTG"
        ret = self.client.confirmarDefinitivo(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosConfirmarDefinitivo=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            establecimiento=establecimiento,
                            codigoCosecha=codigo_cosecha,
                            pesoNeto=peso_neto_carga,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos.get('codigoOperacion', ""))
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def RegresarAOrigenCTGRechazado(self, numero_carta_de_porte, numero_ctg,
                            km_a_recorrer=None,
                            **kwargs):
        "Al consultar los CTGs rechazados se puede Regresar a Origen"
        ret = self.client.regresarAOrigenCTGRechazado(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosRegresarAOrigenCTGRechazado=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg, kmARecorrer=km_a_recorrer,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos['codigoOperacion'])
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def CambiarDestinoDestinatarioCTGRechazado(self, numero_carta_de_porte,
                            numero_ctg, codigo_localidad_destino=None,
                            cuit_destino=None, cuit_destinatario=None,
                            km_a_recorrer=None,
                            **kwargs):
        "Tomar acción de Cambio de Destino y Destinatario para CTG rechazado"
        ret = self.client.cambiarDestinoDestinatarioCTGRechazado(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        datosCambiarDestinoDestinatarioCTGRechazado=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            codigoLocalidadDestino=codigo_localidad_destino,
                            cuitDestino=cuit_destino,
                            cuitDestinatario=cuit_destinatario,
                            kmARecorrer=km_a_recorrer,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('datosResponse')
        if datos:
            self.CartaPorte = str(datos['cartaPorte'])
            self.NumeroCTG = str(datos['ctg'])
            self.FechaHora = str(datos['fechaHora'])
            self.CodigoTransaccion = str(datos['codigoOperacion'])
            self.Observaciones = ""
        return self.CodigoTransaccion

    @inicializar_y_capturar_excepciones
    def ConsultarCTG(self, numero_carta_de_porte=None, numero_ctg=None,
                     patente=None, cuit_solicitante=None, cuit_destino=None,
                     fecha_emision_desde=None, fecha_emision_hasta=None):
        "Operación que realiza consulta de CTGs según el criterio ingresado."
        ret = self.client.consultarCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        consultarCTGDatos=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            patente=patente,
                            cuitSolicitante=cuit_solicitante,
                            cuitDestino=cuit_destino,
                            fechaEmisionDesde=fecha_emision_desde,
                            fechaEmisionHasta=fecha_emision_hasta,
                            )))['response']
        self.__analizar_errores(ret)
        datos = ret.get('arrayDatosConsultarCTG')
        if datos:
            self.DatosCTG = datos
            self.LeerDatosCTG(pop=False)
            return True
        else:
            self.DatosCTG = []
        return ''

    @inicializar_y_capturar_excepciones
    def ConsultarCTGRechazados(self):
        "Consulta de CTGs Otorgados, CTGs Rechazados y CTGs Confirmados"
        ret = self.client.consultarCTGRechazados(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        datos = ret.get('arrayConsultarCTGRechazados')
        if datos:
            self.DatosCTG = datos
            self.LeerDatosCTG(pop=False)
            return True
        else:
            self.DatosCTG = []
        return False

    @inicializar_y_capturar_excepciones
    def ConsultarCTGActivosPorPatente(self, patente="ZZZ999"):
        "Consulta de CTGs activos por patente"
        ret = self.client.consultarCTGActivosPorPatente(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        patente=patente,
                            ))['response']
        self.__analizar_errores(ret)
        datos = ret.get('arrayConsultarCTGActivosPorPatenteResponse')
        if datos:
            self.DatosCTG = datos
            self.LeerDatosCTG(pop=False)
            return True
        else:
            self.DatosCTG = []
        return False

    @inicializar_y_capturar_excepciones
    def CTGsPendientesResolucion(self):
        "Consulta de CTGs Otorgados, CTGs Rechazados y CTGs Confirmados"
        ret = self.client.CTGsPendientesResolucion(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        if ret:
            self.DatosCTG = ret
            return True
        else:
            self.DatosCTG = {}
        return False

    def LeerDatosCTG(self, clave='', pop=True):
        "Recorro los datos devueltos y devuelvo el primero si existe"

        if clave and self.DatosCTG:
            # obtengo la lista por estado pendiente de resolución ("array")
            datos = self.DatosCTG[clave]
        else:
            # uso directamente la lista devuelta por la consulta
            datos = self.DatosCTG
        if datos:
            # extraigo el primer item
            if pop:
                datos = datos.pop(0)
            else:
                datos = datos[0]
            for det in ('datosConsultarCTG', 'detalleConsultaCTGRechazado',
                          'detalleConsultaCTGActivo'):
                if det in datos:
                    datos_ctg = datos[det]
                    break
            else:
                # elemento del array no encontrado:
                return ""
            self.CartaPorte = str(datos_ctg['cartaPorte'])
            self.NumeroCTG = str(datos_ctg['ctg'])
            self.Estado = unicode(datos_ctg.get('estado', ""))
            self.ImprimeConstancia = str(datos_ctg.get('imprimeConstancia', ""))
            for campo in ("fechaRechazo", "fechaEmision", "fechaSolicitud",
                          "fechaConfirmacionArribo"):
                if campo in datos_ctg:
                    self.FechaHora = str(datos_ctg.get(campo))
            self.Destino = datos_ctg.get("destino", "")
            self.Destinatario = datos_ctg.get("destinatario", "")
            self.Observaciones = datos_ctg.get("observaciones", "")
            self.Patente = datos_ctg.get("patente")
            self.PesoNeto = datos_ctg.get("pesoNeto")
            self.FechaVencimiento = datos_ctg.get("fechaVencimiento")
            self.UsuarioSolicitante = datos_ctg.get("usuarioSolicitante")
            self.UsuarioReal = datos_ctg.get("usuarioReal")
            return self.NumeroCTG
        else:
            return ""

    @inicializar_y_capturar_excepciones
    def ConsultarCTGExcel(self, numero_carta_de_porte=None, numero_ctg=None,
                     patente=None, cuit_solicitante=None, cuit_destino=None,
                     fecha_emision_desde=None, fecha_emision_hasta=None,
                     archivo="planilla.xls"):
        "Operación que realiza consulta de CTGs, graba una planilla xls"
        ret = self.client.consultarCTGExcel(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        consultarCTGDatos=dict(
                            cartaPorte=numero_carta_de_porte,
                            ctg=numero_ctg,
                            patente=patente,
                            cuitSolicitante=cuit_solicitante,
                            cuitDestino=cuit_destino,
                            fechaEmisionDesde=fecha_emision_desde,
                            fechaEmisionHasta=fecha_emision_hasta,
                            )))['response']
        self.__analizar_errores(ret)
        datos = base64.b64decode(ret.get('archivo') or "")
        f = open(archivo, "wb")
        f.write(datos)
        f.close()
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarDetalleCTG(self, numero_ctg=None):
        "Operación mostrar este detalle de la  solicitud de CTG seleccionada."
        ret = self.client.consultarDetalleCTG(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        ctg=numero_ctg,
                        ))['response']
        self.__analizar_errores(ret)
        datos = ret.get('consultarDetalleCTGDatos')
        if datos:
            self.NumeroCTG = str(datos['ctg'])
            self.CartaPorte = str(datos['cartaPorte'])
            self.Estado = unicode(datos['estado'])
            self.FechaHora = str(datos['fechaEmision'])
            self.VigenciaDesde = str(datos['fechaVigenciaDesde'])
            self.VigenciaHasta = str(datos['fechaVigenciaHasta'])
            self.TarifaReferencia = str(datos['tarifaReferencia'])
            self.Detalle = str(datos.get('detalle', ""))
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarConstanciaCTGPDF(self, numero_ctg=None,
                                        archivo="constancia.pdf"):
        "Operación Consultar Constancia de CTG en PDF"
        ret = self.client.consultarConstanciaCTGPDF(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        ctg=numero_ctg,
                        ))['response']
        self.__analizar_errores(ret)
        datos = base64.b64decode(ret.get('archivo', ""))
        f = open(archivo, "wb")
        f.write(datos)
        f.close()
        return True

    @inicializar_y_capturar_excepciones
    def ConsultarProvincias(self, sep="||"):
        ret = self.client.consultarProvincias(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['consultarProvinciasResponse']
        self.__analizar_errores(ret)
        array = ret.get('arrayProvincias', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['provincia']['codigo'],
                     it['provincia']['descripcion'])
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarLocalidadesPorProvincia(self, codigo_provincia, sep="||"):
        ret = self.client.consultarLocalidadesPorProvincia(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        codigoProvincia=codigo_provincia,
                        ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayLocalidades', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['localidad']['codigo'],
                     it['localidad']['descripcion'])
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarEstablecimientos(self, sep="||"):
        ret = self.client.consultarEstablecimientos(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                        ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayEstablecimientos', [])
        return [("%s" %
                    (it['establecimiento'],))
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarEspecies(self, sep="||"):
        ret = self.client.consultarEspecies(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayEspecies', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['especie']['codigo'],
                     it['especie']['descripcion'])
               for it in array]

    @inicializar_y_capturar_excepciones
    def ConsultarCosechas(self, sep="||"):
        ret = self.client.consultarCosechas(request=dict(
                        auth={
                            'token': self.Token, 'sign': self.Sign,
                            'cuitRepresentado': self.Cuit, },
                            ))['response']
        self.__analizar_errores(ret)
        array = ret.get('arrayCosechas', [])
        return [("%s %%s %s %%s %s" % (sep, sep, sep)) %
                    (it['cosecha']['codigo'],
                     it['cosecha']['descripcion'])
               for it in array]


def leer_archivo(nombre_archivo):
    archivo = open(nombre_archivo, "r")
    items = []
    ext = os.path.splitext(nombre_archivo)[1]
    if ext == '.csv':
        csv_reader = csv.reader(open(ENTRADA), dialect='excel', delimiter=";")
        for row in csv_reader:
            items.append(row)
        cols = [str(it).strip() for it in items[0]]
        # armar diccionario por cada linea
        items = [dict([(cols[i],str(v).strip()) for i,v in enumerate(item)]) for item in items[1:]]
        return cols, items
    elif ext == '.json':
        items = json.load(archivo)
    elif ext == '.dbf':
        dic = {}
        formatos = [('Encabezado', ENCABEZADO, dic), ]
        leer_dbf(formatos, conf_dbf)
        items = [dic]
    elif ext == '.txt':
        dic = {}
        for linea in archivo:
            if str(linea[0])=='0':
                dic.update(leer(linea, ENCABEZADO))
            else:
                print("Tipo de registro incorrecto:", linea[0])
        items.append(dic)
    else:
        raise RuntimeError("Extension de archivo desconocida: %s" % ext)
    archivo.close()
    cols = [k[0] for k in ENCABEZADO]
    return cols, items


def escribir_archivo(cols, items, nombre_archivo, agrega=False):
    archivo = open(nombre_archivo, agrega and "a" or "w")
    ext = os.path.splitext(nombre_archivo)[1]
    if ext == '.csv':
        csv_writer = csv.writer(archivo, dialect='excel', delimiter=";")
        csv_writer.writerows([cols])
        csv_writer.writerows([[item[k] for k in cols] for item in items])
    elif ext == '.json':
        json.dump(items, archivo, sort_keys=True, indent=4)
    elif ext == '.dbf':
        formatos = [('Encabezado', ENCABEZADO, items), ]
        guardar_dbf(formatos, True, conf_dbf)
    elif ext == '.txt':
        for dic in items:
            dic['tipo_reg'] = 0
            archivo.write(escribir(dic, ENCABEZADO))
    else:
        raise RuntimeError("Extension de archivo desconocida: %s" % ext)
    archivo.close()



# busco el directorio de instalación (global para que no cambie si usan otra dll)
if not hasattr(sys, "frozen"):
    basepath = __file__
elif sys.frozen=='dll':
    import win32api
    basepath = win32api.GetModuleFileName(sys.frozendllhandle)
else:
    basepath = sys.executable
INSTALL_DIR = WSCTGv2.InstallDir = get_install_dir()


if __name__ == '__main__':
    if '--ayuda' in sys.argv:
        print(LICENCIA)
        print(AYUDA)
        sys.exit(0)
    if '--formato' in sys.argv:
        print("Formato:")
        for msg, formato in [('Encabezado', ENCABEZADO), ]:
            comienzo = 1
            print("=== %s ===" % msg)
            for fmt in formato:
                clave, longitud, tipo = fmt[0:3]
                dec = len(fmt)>3 and fmt[3] or (tipo=='I' and '2' or '')
                print(" * Campo: %-20s Posición: %3d Longitud: %4d Tipo: %s Decimales: %s" % (
                    clave, comienzo, longitud, tipo, dec))
                comienzo += longitud
        sys.exit(0)

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSCTGv2)
        sys.exit(0)

    import csv
    from ConfigParser import SafeConfigParser

    try:

        if "--version" in sys.argv:
            print("Versión: ", __version__)

        for arg in sys.argv[1:]:
            if arg.startswith("--"):
                break
            print("Usando configuración:", arg)
            CONFIG_FILE = arg

        config = SafeConfigParser()
        config.read(CONFIG_FILE)
        CERT = config.get('WSAA','CERT')
        PRIVATEKEY = config.get('WSAA','PRIVATEKEY')
        CUIT = config.get('WSCTG','CUIT')
        ENTRADA = config.get('WSCTG','ENTRADA')
        SALIDA = config.get('WSCTG','SALIDA')

        if config.has_option('WSAA','URL') and not HOMO:
            wsaa_url = config.get('WSAA','URL')
        else:
            wsaa_url = None
        if config.has_option('WSCTG','URL') and not HOMO:
            wsctg_url = config.get('WSCTG','URL')
        else:
            wsctg_url = WSDL

        if config.has_section('DBF'):
            conf_dbf = dict(config.items('DBF'))
            if DEBUG:
                print("conf_dbf", conf_dbf)
        else:
            conf_dbf = {}

        DEBUG = '--debug' in sys.argv
        XML = '--xml' in sys.argv

        if DEBUG:
            print("Usando Configuración:")
            print("wsaa_url:", wsaa_url)
            print("wsctg_url:", wsctg_url)

        # obteniendo el TA
        from wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wsctg", CERT, PRIVATEKEY, wsaa_url, debug=DEBUG)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)

        # cliente soap del web service
        wsctg = WSCTGv2()
        wsctg.Conectar(wsdl=wsctg_url)
        wsctg.SetTicketAcceso(ta)
        wsctg.Cuit = CUIT

        if '--dummy' in sys.argv:
            ret = wsctg.Dummy()
            print("AppServerStatus", wsctg.AppServerStatus)
            print("DbServerStatus", wsctg.DbServerStatus)
            print("AuthServerStatus", wsctg.AuthServerStatus)
            sys.exit(0)

        if '--anular' in sys.argv:
            i = sys.argv.index("--anular")
            # print(wsctg.client.help("anularCTG"))
            if i + 2 > len(sys.argv) or sys.argv[i + 1].startswith("--"):
                carta_porte = raw_input("Ingrese Carta de Porte: ")
                ctg = raw_input("Ingrese CTG: ")
            else:
                carta_porte = sys.argv[i + 1]
                ctg = sys.argv[i + 2]
            ret = wsctg.AnularCTG(carta_porte, ctg)
            wsctg.SolicitarCTGDatoPendiente()
            print("Carta Porte", wsctg.CartaPorte)
            print("Numero CTG", wsctg.NumeroCTG)
            print("Fecha y Hora", wsctg.FechaHora)
            print("Codigo Anulacion de CTG", wsctg.CodigoOperacion)
            print("Errores:", wsctg.Errores)
            sys.exit(0)

        if '--rechazar' in sys.argv:
            i = sys.argv.index("--rechazar")
            # print(wsctg.client.help("rechazarCTG"))
            if i + 3 > len(sys.argv) or sys.argv[i + 1].startswith("--"):
                carta_porte = raw_input("Ingrese Carta de Porte: ")
                ctg = raw_input("Ingrese CTG: ")
                motivo = raw_input("Motivo: ")
            else:
                carta_porte = sys.argv[i + 1]
                ctg = sys.argv[i + 2]
                motivo = sys.argv[i + 3]
            ret = wsctg.RechazarCTG(carta_porte, ctg, motivo)
            print("Carta Porte", wsctg.CartaPorte)
            print("Numero CTG", wsctg.NumeroCTG)
            print("Fecha y Hora", wsctg.FechaHora)
            print("Codigo Anulacion de CTG", wsctg.CodigoOperacion)
            print("Errores:", wsctg.Errores)
            sys.exit(0)


        # Recuperar parámetros:

        if '--provincias' in sys.argv:
            ret = wsctg.ConsultarProvincias()
            print("\n".join(ret))

        if '--localidades' in sys.argv:
            ret = wsctg.ConsultarLocalidadesPorProvincia(16)
            print("\n".join(ret))

        if '--especies' in sys.argv:
            ret = wsctg.ConsultarEspecies()
            print("\n".join(ret))

        if '--cosechas' in sys.argv:
            ret = wsctg.ConsultarCosechas()
            print("\n".join(ret))

        if '--establecimientos' in sys.argv:
            ret = wsctg.ConsultarEstablecimientos()
            print("\n".join(ret))


        if '--prueba' in sys.argv or '--formato' in sys.argv:
            prueba = dict(numero_carta_de_porte=512345679, codigo_especie=23,
                cuit_canjeador=0, #30660685908,
                cuit_destino=20111111112, cuit_destinatario=20222222223,
                codigo_localidad_origen=3058, codigo_localidad_destino=3059,
                codigo_cosecha='1314', peso_neto_carga=1000,
                km_a_recorrer=1234,
                observaciones='', establecimiento=1,
            )
            if [argv for argv in sys.argv if argv.startswith(("--confirmar",
                    "--regresar", '--cambiar'))]:
                prueba.update(dict(
                    numero_ctg="49241727", transaccion='10000001681',
                    consumo_propio='S',
                    ))
            parcial = dict(
                    cant_horas=1,
                    patente_vehiculo='APE652', cuit_transportista=20333333334,
                    )
            if not '--parcial' in sys.argv:
                prueba.update(parcial)

            escribir_archivo(prueba.keys(), [prueba], ENTRADA)

        cols, items = leer_archivo(ENTRADA)
        ctg = None

        if '--solicitar' in sys.argv:
            wsctg.LanzarExcepciones = True
            for it in items:
                print("solicitando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                ctg = wsctg.SolicitarCTGInicial(**it)
                print("numero CTG: ", ctg)
                print("Observiacion: ", wsctg.Observaciones)
                print("Carta Porte", wsctg.CartaPorte)
                print("Numero CTG", wsctg.NumeroCTG)
                print("Fecha y Hora", wsctg.FechaHora)
                print("Vigencia Desde", wsctg.VigenciaDesde)
                print("Vigencia Hasta", wsctg.VigenciaHasta)
                print("Tarifa Referencia: ", wsctg.TarifaReferencia)
                print("Errores:", wsctg.Errores)
                print("Controles:", wsctg.Controles)
                it['numero_ctg'] = wsctg.NumeroCTG
                it['tarifa_referencia'] = wsctg.TarifaReferencia
                it['observaciones'] = wsctg.Observaciones
                it['fecha_hora'] = wsctg.FechaHora
                it['vigencia_desde'] = wsctg.VigenciaDesde
                it['vigencia_hasta'] = wsctg.VigenciaHasta
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Controles)

        if '--parcial' in sys.argv:
            wsctg.LanzarExcepciones = True
            for it in items:
                print("solicitando dato pendiente...", ' '.join(['%s=%s' % (k,v) for k,v in parcial.items()]))
                ctg = wsctg.SolicitarCTGDatoPendiente(
                    numero_carta_de_porte=wsctg.CartaPorte,
                    **parcial)
                print("numero CTG: ", ctg)
                print("Observiacion: ", wsctg.Observaciones)
                print("Carta Porte", wsctg.CartaPorte)
                print("Numero CTG", wsctg.NumeroCTG)
                print("Fecha y Hora", wsctg.FechaHora)
                print("Vigencia Desde", wsctg.VigenciaDesde)
                print("Vigencia Hasta", wsctg.VigenciaHasta)
                print("Tarifa Referencia: ", wsctg.TarifaReferencia)
                print("Errores:", wsctg.Errores)
                print("Controles:", wsctg.Controles)
                it['numero_ctg'] = wsctg.NumeroCTG
                it['tarifa_referencia'] = wsctg.TarifaReferencia
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Controles)

        if '--confirmar_arribo' in sys.argv:
            for it in items:
                print("confirmando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.ConfirmarArribo(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Controles)

        if '--confirmar_definitivo' in sys.argv:
            if '--testing' in sys.argv:
                wsctg.LoadTestXML("wsctg_confirmar_def.xml") # cargo respuesta
            for it in items:
                print("confirmando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.ConfirmarDefinitivo(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Errores)

        if '--regresar_a_origen_rechazado' in sys.argv:
            for it in items:
                print("regresando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.RegresarAOrigenCTGRechazado(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Errores)

        if '--cambiar_destino_destinatario_rechazado' in sys.argv:
            for it in items:
                print("cambiando...", ' '.join(['%s=%s' % (k,v) for k,v in it.items()]))
                transaccion = wsctg.CambiarDestinoDestinatarioCTGRechazado(**it)
                print("transaccion: %s" % (transaccion, ))
                print("Fecha y Hora", wsctg.FechaHora)
                print("Errores:", wsctg.Errores)
                it['transaccion'] = transaccion
                it['errores'] = '|'.join(wsctg.Errores)
                it['controles'] = '|'.join(wsctg.Errores)

        if '--consultar_detalle' in sys.argv:
            i = sys.argv.index("--consultar_detalle")
            if len(sys.argv) > i + 1 and not sys.argv[i+1].startswith("--"):
                ctg = int(sys.argv[i+1])
            elif not ctg:
                ctg = int(raw_input("Numero de CTG: ") or '0') or 73714620

            wsctg.LanzarExcepciones = True
            for i, it in enumerate(items):
                print("consultando detalle...", ctg)
                ok = wsctg.ConsultarDetalleCTG(ctg)
                print("Numero CTG: ", wsctg.NumeroCTG)
                print("Tarifa Referencia: ", wsctg.TarifaReferencia)
                print("Observiacion: ", wsctg.Observaciones)
                print("Carta Porte", wsctg.CartaPorte)
                print("Numero CTG", wsctg.NumeroCTG)
                print("Fecha y Hora", wsctg.FechaHora)
                print("Vigencia Desde", wsctg.VigenciaDesde)
                print("Vigencia Hasta", wsctg.VigenciaHasta)
                print("Errores:", wsctg.Errores)
                print("Controles:", wsctg.Controles)
                print("Detalle:", wsctg.Detalle)
                it['numero_ctg'] = wsctg.NumeroCTG
                it['observaciones'] = wsctg.Observaciones
                it['fecha_hora'] = wsctg.FechaHora
                it['vigencia_desde'] = wsctg.VigenciaDesde
                it['vigencia_hasta'] = wsctg.VigenciaHasta
                wsctg.AnalizarXml("XmlResponse")
                for k, ki in {'ctg': 'numero_ctg', 'solicitante': '',
                          'estado': 'estado',
                          'especie': '', ##'codigo_especie', no devuelve codigo!
                          'cosecha': '', ##'codigo_cosecha', no devuelve codigo!
                          'cuitCanjeador': 'cuit_canjeador',
                          'cuitDestino': 'cuit_destino',
                          'cuitDestinatario': 'cuit_destinatario',
                          'cuitTransportista': 'cuit_transportista',
                          'establecimiento': 'establecimiento',
                          'localidadOrigen': 'localidad_origen',
                          'localidadDestino': 'localidad_destino',
                          'cantidadHoras': 'cantidad_horas',
                          'patenteVehiculo': 'patente_vehiculo',
                          'pesoNetoCarga': 'peso_neto_carga',
                          'kmRecorridos': 'km_recorridos',
                          'tarifaReferencia': 'tarifa_referencia', }.items():
                    v = wsctg.ObtenerTagXml('consultarDetalleCTGDatos', k)
                    print(k, v)
                    if ki.startswith("cuit") and v:
                        v = v[:11]
                    it[ki] = v

        escribir_archivo(cols, items, SALIDA)

        if "--consultar" in sys.argv:
            wsctg.LanzarExcepciones = True
            wsctg.ConsultarCTG(fecha_emision_desde="01/04/2012")
            print("Numero CTG - Carta de Porte - Imprime Constancia - Estado")
            while wsctg.LeerDatosCTG():
                print(wsctg.NumeroCTG, wsctg.CartaPorte,)
                print(wsctg.ImprimeConstancia, wsctg.Estado, wsctg.FechaHora)

        if "--consultar_rechazados" in sys.argv:
            wsctg.LanzarExcepciones = True
            wsctg.ConsultarCTGRechazados()
            print("Numero CTG - Carta de Porte - Fecha - Destino/Dest./Obs.")
            while wsctg.LeerDatosCTG():
                print(wsctg.NumeroCTG, wsctg.CartaPorte, wsctg.FechaHora,)
                print(wsctg.Destino, wsctg.Destinatario, wstcg.Observaciones)

        if "--consultar_activos_por_patente" in sys.argv:
            i = sys.argv.index("--consultar_activos_por_patente")
            if len(sys.argv) > i + 1 and not sys.argv[i+1].startswith("--"):
                patente = int(sys.argv[i+1])
            elif not ctg:
                patente= raw_input("Patente: ") or 'APE652'
            wsctg.LanzarExcepciones = True
            if '--testing' in sys.argv:
                wsctg.LoadTestXML("wsctgv2_activos.xml")
            wsctg.ConsultarCTGActivosPorPatente(patente=patente)
            print("Numero CTG - Carta de Porte - Fecha - Peso Neto - Usuario")
            while wsctg.LeerDatosCTG():
                print(wsctg.NumeroCTG, wsctg.CartaPorte, wsctg.Patente,)
                print(wsctg.FechaHora, wsctg.FechaVencimiento, wsctg.PesoNeto,)
                print(wsctg.UsuarioSolicitante, wsctg.UsuarioReal)

        if '--consultar_excel' in sys.argv:
            archivo = raw_input("Archivo a generar (planilla.xls): ") or \
                        'planilla.xls'
            wsctg.LanzarExcepciones = True
            ok = wsctg.ConsultarCTGExcel(fecha_emision_desde="01/04/2012",
                                         archivo=archivo)
            print("Errores:", wsctg.Errores)

        if '--consultar_constancia_pdf' in sys.argv:
            i = sys.argv.index("--consultar_constancia_pdf")
            if len(sys.argv) > i + 2 and not sys.argv[i+1].startswith("--"):
                ctg = int(sys.argv[i+1])
                archivo = sys.argv[i+2]
            elif not ctg:
                ctg = int(raw_input("Numero de CTG: ") or '0') or 83139794
                archivo = raw_input("Archivo a generar (constancia.pdf): ") or \
                            'constancia.pdf'

            wsctg.LanzarExcepciones = True
            ok = wsctg.ConsultarConstanciaCTGPDF(ctg, archivo)
            print("Errores:", wsctg.Errores)

        if "--pendientes" in sys.argv:
            wsctg.LanzarExcepciones = True
            wsctg.CTGsPendientesResolucion()
            for clave in ("arrayCTGsRechazadosAResolver",
                          "arrayCTGsOtorgadosAResolver",
                          "arrayCTGsConfirmadosAResolver", ):
                print(clave[6:])
                print("Numero CTG - Carta de Porte - Imprime Constancia - Estado")
                while wsctg.LeerDatosCTG(clave):
                    print(wsctg.NumeroCTG, wsctg.CartaPorte, wsctg.FechaHora)
                    print(wsctg.Destino, wsctg.Destinatario, wsctg.Observaciones)

        print("hecho.")

    except SoapFault as e:
        print("Falla SOAP:", e.faultcode, e.faultstring.encode("ascii","ignore"))
        sys.exit(3)
    except Exception as e:
        ex = utils.exception_info()
        print(ex)
        if DEBUG:
            raise
        sys.exit(5)



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Módulo de Intefase para archivos de texto (exportación version 1)"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2011 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.27b"

import datetime
import os
import sys
import time
# import traceback
from ConfigParser import SafeConfigParser

# revisar la instalación de pyafip.ws:
import wsfexv1
# from php import (
#     SimpleXMLElement,
#     SoapClient,
#     SoapFault,
#     date
# )
from utils import leer, escribir, leer_dbf, guardar_dbf, N, A, I


HOMO = wsfexv1.HOMO
DEBUG = False
XML = False
CONFIG_FILE = "rece.ini"

LICENCIA = """
recex.py: Interfaz de texto para generar Facturas Electrónica Exportación
Copyright (C) 2010 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional sobre garantía, soporte técnico comercial
e incorporación/distribución en programas propietarios ver PyAfipWs:
http://www.sistemasagiles.com.ar/trac/wiki/PyAfipWs
"""

# definición del formato del archivo de intercambio:

if '--pyfepdf' not in sys.argv:
    TIPOS_REG = '0', '1', '2', '3'
    ENCABEZADO = [
        ('tipo_reg', 1, N),  # 0: encabezado
        ('fecha_cbte', 8, A),
        ('tipo_cbte', 2, N),
        ('punto_vta', 4, N),
        ('cbte_nro', 8, N),
        ('tipo_expo', 1, N),  # 1:bienes, 2:servicios,...
        ('permiso_existente', 1, A),  # S/N/
        ('pais_dst_cmp', 3, N),  # 203
        ('nombre_cliente', 200, A),  # 'Joao Da Silva'
        ('cuit_pais_cliente', 11, N),  # 50000000016
        ('domicilio_cliente', 300, A),  # 'Rua 76 km 34.5 Alagoas'
        ('id_impositivo', 50, A),  # 'PJ54482221-l'
        ('imp_total', 15, I, 2),
        ('moneda_id', 3, A),
        ('moneda_ctz', 10, I, 6),  # 10,6
        ('obs_comerciales', 4000, A),
        ('obs_generales', 1000, A),
        ('forma_pago', 50, A),
        ('incoterms', 3, A),
        ('incoterms_ds', 20, A),
        ('idioma_cbte', 1, A),
        ('cae', 14, N), ('fecha_vto', 8, A),
        ('resultado', 1, A),
        ('reproceso', 1, A),
        ('motivos_obs', 1000, A),
        ('id', 15, N),
        ('fch_venc_cae', 8, A),
        ('excepcion', 100, A),
        ('err_code', 100, A),
        ('err_msg', 1000, A),
        ]

    DETALLE = [
        ('tipo_reg', 1, N),  # 1: detalle item
        ('codigo', 50, A),
        ('qty', 12, I, 6),
        ('umed', 2, N),
        ('precio', 12, I, 6),
        ('importe', 13, I, 2),
        ('bonif', 12, I, 6),
        ('ds', 4000, A),
        ]

    PERMISO = [
        ('tipo_reg', 1, N),  # 2: permiso
        ('id_permiso', 16, A),
        ('dst_merc', 3, N),
        ]

    CMP_ASOC = [
        ('tipo_reg', 1, N),  # 3: comprobante asociado
        ('cbte_tipo', 3, N), ('cbte_punto_vta', 4, N),
        ('cbte_nro', 8, N), ('cbte_cuit', 11, N),
        ]
else:
    print("!" * 78)
    print("importando formato segun pyfepdf")
    from formato_txt import (
        ENCABEZADO,
        DETALLE,
        PERMISO,
        CMP_ASOC,
        # IVA,
        # TRIBUTO
    )
    TIPOS_REG = '0', '1', '2', '3'

if '/recex' in sys.argv:
    from recex import (
        ENCABEZADO,
        DETALLE,
        PERMISO,
        CMP_ASOC
    )
    ENCABEZADO[8] = ('nombre_cliente', 200, A)  # 'Joao Da Silva'
    ENCABEZADO[7] = ('pais_dst_cmp', 3, N)
    ENCABEZADO[16] = ('obs_generales', 1000, A)
    DETALLE[5] = ('importe', 13, I, 2)
    DETALLE.append(('bonif', 12, I, 6))


def autorizar(ws, entrada, salida):
    # recupero el último número de transacción
    # id = wsfex.ultnro(client, token, sign, cuit)

    detalles = []
    permisos = []
    cbtasocs = []
    encabezado = []
    if '/dbf' in sys.argv:
        formatos = [
            ('Encabezado', ENCABEZADO, encabezado),
            ('Permisos', PERMISO, permisos),
            ('Comprobante Asociado', CMP_ASOC, cbtasocs),
            ('Detalles', DETALLE, detalles)]
        dic = leer_dbf(formatos, conf_dbf)
        encabezado = encabezado[0]
    else:
        encabezado = {}
        for linea in entrada:
            if str(linea[0]) == TIPOS_REG[0]:
                encabezado = leer(linea, ENCABEZADO)
                if 'nro_doc' in encabezado:
                    encabezado['cuit_pais_cliente'] = encabezado['nro_doc']
            elif str(linea[0]) == TIPOS_REG[1]:
                detalle = leer(linea, DETALLE)
                detalles.append(detalle)
            elif str(linea[0]) == TIPOS_REG[2]:
                permiso = leer(linea, PERMISO)
                permisos.append(permiso)
            elif str(linea[0]) == TIPOS_REG[3]:
                cbtasoc = leer(linea, CMP_ASOC)
                cbtasocs.append(cbtasoc)
            else:
                print("Tipo de registro incorrecto:", linea[0])

    if not encabezado['id']:
        # TODO: habria que leer y/o grabar el id en el archivo
        # id += 1 # incremento el nº de transacción
        # Por el momento, el id se calcula con el tipo, pv y nº de comprobant
        i = long(encabezado['cbte_nro'])
        i += (
            int(encabezado['cbte_nro'])*10**4 +
            int(encabezado['punto_vta']))*10**8
        encabezado['id'] = ws.GetLastID() + 1

    if '/testing' in sys.argv:
        encabezado['id'] = long(ws.GetLastID()) + 1
        encabezado['cbte_nro'] = int(
            ws.GetLastCMP(encabezado['tipo_cbte'], encabezado['punto_vta'])
        ) + 1
        encabezado['fecha_cbte'] = datetime.datetime.now().strftime("%Y%m%d")

    ws.CrearFactura(**encabezado)
    for detalle in detalles:
        ws.AgregarItem(**detalle)
    for permiso in permisos:
        ws.AgregarPermiso(**permiso)
    for cbtasoc in cbtasocs:
        ws.AgregarCmpAsoc(**cbtasoc)

    if DEBUG:
        # print(f.to_dict())
        print(
            '\n'.join(
                ["%s='%s'" % (k, str(v)) for k, v in encabezado.items()]))
        for detalle in detalles:
            print(
                ', '.join(
                    ["%s='%s'" % (k, str(v)) for k, v in detalle.items()]))
            print(
                "DIF:", detalle['qty'] * detalle['precio'] - detalle['importe']
            )

        print('id:', encabezado['id'])
    if not DEBUG or not sys.stdout.isatty() or raw_input("Facturar?") == "S":
        ws.LanzarExcepcion = False
        cae = ws.Authorize(id=encabezado['id'])
        dic = ws.factura
        dic.update({
            'cae': cae and str(cae) or '',
            'fch_venc_cae': ws.FchVencCAE and str(ws.FchVencCAE) or '',
            'resultado': ws.Resultado or '',
            'motivos_obs': ws.Obs or '',
            'err_code': str(ws.ErrCode),
            'err_msg': ws.ErrMsg or '',
            'reproceso': ws.Reproceso or '',
            })
        escribir_factura(dic, salida)
        print(
            "ID:", encabezado['id'], "NRO:", dic['cbte_nro'],
            "Resultado:", dic['resultado'],
        )
        print(
            "CAE:", dic['cae'], "Obs:",
            dic['motivos_obs'].encode("ascii", "ignore"),
        )
        print(
            "Err:", dic['err_msg'].encode("ascii", "ignore"),
            "Reproceso:", dic['reproceso'])
        if ws.Excepcion:
            print("Excepcion:", ws.Excepcion.encode("ascii", "ignore"))
            print("Traceback:", ws.Traceback.encode("ascii", "ignore"))


def escribir_factura(dic, archivo, agrega=False):
    dic['tipo_reg'] = TIPOS_REG[0]
    archivo.write(escribir(dic, ENCABEZADO))
    for it in dic.get('detalles', []):
        it['tipo_reg'] = TIPOS_REG[1]
        archivo.write(escribir(it, DETALLE))
    if 'permisos' in dic:
        for it in dic['permisos']:
            it['tipo_reg'] = TIPOS_REG[2]
            archivo.write(escribir(it, PERMISO))

    if '/dbf' in sys.argv:
        formatos = [
            ('Encabezado', ENCABEZADO, [dic]),
            ('Permisos', PERMISO, dic.get('permisos', [])),
            ('Comprobante Asociado', CMP_ASOC, dic.get('cbtes_asoc', [])),
            ('Detalles', DETALLE, dic.get('detalles', []))
        ]
        guardar_dbf(formatos, agrega, conf_dbf)


def depurar_xml(client):
    fecha = time.strftime("%Y%m%d%H%M%S")
    f = open("request-%s.xml" % fecha, "w")
    f.write(client.xml_request)
    f.close()
    f = open("response-%s.xml" % fecha, "w")
    f.write(client.xml_response)
    f.close()


if __name__ == "__main__":
    if '/ayuda' in sys.argv:
        print(LICENCIA)
        print("Opciones: ")
        print(" /ayuda: este mensaje")
        print(" /dummy: consulta estado de servidores")
        print(
            " /prueba: genera y autoriza una factura de prueba \
            (no usar en producción!)"
        )
        print(" /ult: consulta último número de comprobante")
        print(" /debug: modo depuración (detalla y confirma las operaciones)")
        print(
            " /formato: muestra el formato de los archivos de entrada/salida")
        print(
            " /get: recupera datos de un comprobante autorizado previamente \
            (verificación)"
        )
        print(
            " /xml: almacena los requerimientos y respuestas XML (depuración)"
        )
        print(" /dbf: lee y almacena la información en tablas DBF")
        print(
            "Ver rece.ini para parámetros de configuración \
            (URL, certificados, etc.)"
        )
        sys.exit(0)

    # si se pasa el archivo de configuración por parámetro, confirmar que
    # exista
    # y descartar que sea una opción
    if len(sys.argv) > 1 and \
       (sys.argv[1][0] not in "-/" or os.path.exists(sys.argv[1])):
        CONFIG_FILE = sys.argv.pop(1)
    if DEBUG:
        print("CONFIG_FILE:", CONFIG_FILE)

    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    cert = config.get('WSAA', 'CERT')
    privatekey = config.get('WSAA', 'PRIVATEKEY')
    cuit = config.get('WSFEXv1', 'CUIT')
    entrada = config.get('WSFEXv1', 'ENTRADA')
    salida = config.get('WSFEXv1', 'SALIDA')

    if config.has_option('WSAA', 'URL') and not HOMO:
        wsaa_url = config.get('WSAA', 'URL')
    else:
        wsaa_url = None
    if config.has_option('WSFEXv1', 'URL') and not HOMO:
        wsfexv1_url = config.get('WSFEXv1', 'URL')
    else:
        wsfexv1_url = ""

    if config.has_section('DBF'):
        conf_dbf = dict(config.items('DBF'))
        if DEBUG:
            print("conf_dbf", conf_dbf)
    else:
        conf_dbf = {}

    if '/debug'in sys.argv:
        DEBUG = True

    if '/xml'in sys.argv:
        XML = True

    if DEBUG:
        print("wsaa_url %s\nwsfexv1_url %s" % (wsaa_url, wsfexv1_url))
        print("Config_file:", CONFIG_FILE)
        print("Entrada: ", entrada)
        print("Salida:", salida)

    try:
        ws = wsfexv1.WSFEXv1()
        ws.Conectar("", wsfexv1_url)
        ws.Cuit = cuit

        if '/dummy' in sys.argv:
            print("Consultando estado de servidores...")
            ws.Dummy()
            print("AppServerStatus", ws.AppServerStatus)
            print("DbServerStatus", ws.DbServerStatus)
            print("AuthServerStatus", ws.AuthServerStatus)
            sys.exit(0)

        if '/formato' in sys.argv:
            from formatos.formato_dbf import definir_campos
            print("Formato:")
            for msg, formato in [
                ('Encabezado', ENCABEZADO),
                ('Detalle', DETALLE),
                ('Permiso', PERMISO),
                ('Comprobante Asociado', CMP_ASOC)
            ]:
                if '/dbf' not in sys.argv:
                    comienzo = 1
                    print("== %s ==" % msg)
                    for fmt in formato:
                        clave, longitud, tipo = fmt[0:3]
                        dec = len(fmt) > 3 and fmt[3] or (
                            tipo == 'I' and '2' or '')
                        print(
                            " * Campo: %-20s Posición: %3d Longitud: %4d \
                            Tipo: %s Decimales: %s" % (
                                clave, comienzo, longitud, tipo, dec))
                        comienzo += longitud
                else:
                    filename = "%s.dbf" % msg.lower()[:8]
                    print("==== %s (%s) ====" % (msg, filename))
                    claves, campos = definir_campos(formato)
                    for campo in campos:
                        print(" * Campo: %s" % (campo,))
            sys.exit(0)

        # obteniendo el TA
        from wsaa import WSAA
        wsaa = WSAA()
        ta = wsaa.Autenticar("wsfex", cert, privatekey, wsaa_url)
        if not ta:
            sys.exit("Imposible autenticar con WSAA: %s" % wsaa.Excepcion)
        ws.SetTicketAcceso(ta)

        if '/prueba' in sys.argv:
            # generar el archivo de prueba para la próxima factura
            f_entrada = open(entrada, "w")

            tipo_cbte = 19  # FC Expo (ver tabla de parámetros)
            punto_vta = 7
            # Obtengo el último número de comprobante y le agrego 1
            cbte_nro = int(ws.GetLastCMP(tipo_cbte, punto_vta)) + 1
            fecha_cbte = datetime.datetime.now().strftime("%Y%m%d")
            tipo_expo = 1  # tipo de exportación (ver tabla de parámetros)
            permiso_existente = "S"
            dst_cmp = 203  # país destino
            cliente = "Joao Da Silva"
            cuit_pais_cliente = "50000000016"
            domicilio_cliente = "Rua 76 km 34.5 Alagoas"
            id_impositivo = "PJ54482221-l"
            # para reales, "DOL" o "PES" (ver tabla de parámetros)
            moneda_id = "DOL"
            moneda_ctz = 8.02
            obs_comerciales = "Observaciones comerciales"
            obs = "Sin observaciones"
            forma_pago = "30 dias"
            incoterms = "FOB"  # (ver tabla de parámetros)
            incoterms_ds = "Flete a Bordo"
            idioma_cbte = 1  # (ver tabla de parámetros)
            imp_total = "250.00"

            # Creo una factura (internamente, no se llama al WebService):
            ok = ws.CrearFactura(
                tipo_cbte, punto_vta, cbte_nro, fecha_cbte,
                imp_total, tipo_expo, permiso_existente, dst_cmp,
                cliente, cuit_pais_cliente, domicilio_cliente,
                id_impositivo, moneda_id, moneda_ctz,
                obs_comerciales, obs, forma_pago, incoterms,
                idioma_cbte, incoterms_ds)

            # Agrego un item:
            codigo = "PRO1"
            ds = "Producto Tipo 1 Exportacion MERCOSUR ISO 9001"
            qty = 2
            precio = "150.00"
            umed = 1  # Ver tabla de parámetros (unidades de medida)
            bonif = "50.00"
            imp_total = "250.00"  # importe total final del artículo
            # lo agrego a la factura (internamente, no se llama al WebService):
            ok = ws.AgregarItem(
                codigo, ds, qty, umed, precio, imp_total, bonif)

            # Agrego un permiso (ver manual para el desarrollador)
            id = "99999AAXX999999A"
            dst = 225  # país destino de la mercaderia
            ok = ws.AgregarPermiso(id, dst)

            # Agrego un comprobante asociado (solo para N/C o N/D)
            if tipo_cbte in (20, 21):
                cbteasoc_tipo = 19
                cbteasoc_pto_vta = 2
                cbteasoc_nro = 1234
                cbteasoc_cuit = 20111111111
                ws.AgregarCmpAsoc(
                    cbteasoc_tipo, cbteasoc_pto_vta, cbteasoc_nro,
                    cbteasoc_cuit)

            dic = ws.factura
            dic['id'] = ws.GetLastID() + 1
            escribir_factura(dic, f_entrada, agrega=True)
            f_entrada.close()

        if '/ult' in sys.argv:
            i = sys.argv.index("/ult")
            if i+2 < len(sys.argv):
                tipo_cbte = int(sys.argv[i+1])
                punto_vta = int(sys.argv[i+2])
            else:
                tipo_cbte = int(raw_input("Tipo de comprobante: "))
                punto_vta = int(raw_input("Punto de venta: "))
            ult_cbte = ws.GetLastCMP(tipo_cbte, punto_vta)
            print("Ultimo numero: ", ult_cbte)
            print(ws.ErrMsg)
            depurar_xml(ws.client)
            escribir_factura({'tipo_cbte': tipo_cbte,
                              'punto_vta': punto_vta,
                              'cbte_nro': ult_cbte,
                              'fecha_cbte': ws.FechaCbte,
                              'err_msg': ws.ErrMsg,
                              }, open(salida, "w"))
            sys.exit(0)

        if '/get' in sys.argv:
            print("Recuperar comprobante:")
            i = sys.argv.index("/get")
            if i+3 < len(sys.argv):
                tipo_cbte = int(sys.argv[i+1])
                punto_vta = int(sys.argv[i+2])
                cbte_nro = int(sys.argv[i+3])
            else:
                tipo_cbte = int(raw_input("Tipo de comprobante: "))
                punto_vta = int(raw_input("Punto de venta: "))
                cbte_nro = int(raw_input("Numero de comprobante: "))
            ws.GetCMP(tipo_cbte, punto_vta, cbte_nro)

            print("FechaCbte = ", ws.FechaCbte)
            print("CbteNro = ", ws.CbteNro)
            print("PuntoVenta = ", ws.PuntoVenta)
            print("ImpTotal =", ws.ImpTotal)
            print("CAE = ", ws.CAE)
            print("Vencimiento = ", ws.Vencimiento)
            print(ws.ErrMsg)

            depurar_xml(ws.client)
            escribir_factura(
                {'tipo_cbte': tipo_cbte,
                 'punto_vta': ws.PuntoVenta,
                 'cbte_nro': ws.CbteNro,
                 'fecha_cbte': ws.FechaCbte,
                 'imp_total': ws.ImpTotal,
                 'cae': str(ws.CAE),
                 'fch_venc_cae': ws.Vencimiento,
                 'err_msg': ws.ErrMsg},
                open(salida, "w")
            )
            sys.exit(0)

        if '/ctz' in sys.argv:
            i = sys.argv.index("/ctz")
            if i+1 < len(sys.argv):
                moneda_id = sys.argv[i+1]
            else:
                moneda_id = raw_input("Id de moneda (DOL): ") or 'DOL'
            ctz = ws.GetParamCtz(moneda_id)
            print("Cotizacion: ", ctz)
            print(ws.ErrMsg)
            sys.exit(0)

        f_entrada = f_salida = None
        try:
            f_entrada = open(entrada, "r")
            f_salida = open(salida, "w")
            try:
                autorizar(ws, f_entrada, f_salida)
            except Exception as e:
                XML = True
                raise
        finally:
            if f_entrada is not None:
                f_entrada.close()
            if f_salida is not None:
                f_salida.close()
            if XML:
                depurar_xml(ws.client)
        sys.exit(0)

    except Exception as e:
        print(unicode(e).encode("ascii", "ignore"))
        if DEBUG or True:
            raise
        sys.exit(5)



#!usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"Aplicativo AdHoc Para generación de Facturas Electrónicas"

__author__ = "Mariano Reingart (reingart@gmail.com)"
__copyright__ = "Copyright (C) 2009-2015 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.27f"

from datetime import datetime
from decimal import Decimal, getcontext, ROUND_DOWN
import os
import sys
import wx
import gui
import unicodedata
import traceback
from ConfigParser import SafeConfigParser
import wsaa, wsfe, wsfev1, wsfexv1
from php import SimpleXMLElement, SoapClient, SoapFault, date
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from smtplib import SMTP

# from PyFPDF.ejemplos.form import Form
try:
    from pyfepdf import FEPDF
except Exception as e:
    print(e)

# Formatos de archivos:
from formatos import formato_xml, formato_csv, formato_dbf, formato_txt, formato_json

HOMO = False
DEBUG = '--debug' in sys.argv
CONFIG_FILE = "rece.ini"

ACERCA_DE = u"""
PyRece: Aplicativo AdHoc para generar Facturas Electrónicas
Copyright (C) 2008-2015 Mariano Reingart reingart@gmail.com

Este progarma es software libre, se entrega ABSOLUTAMENTE SIN GARANTIA
y es bienvenido a redistribuirlo bajo la licencia GPLv3.

Para información adicional y descargas ver:
http://www.sistemasagiles.com.ar/
"""

INSTRUCTIVO = U"""
Forma de uso:

 * Examinar: para buscar el archivo a procesar (opcional)
 * Cargar: para leer los datos del archivo de facturas a procesar
 * Autenticar: para iniciar la sesión en los servidores de AFIP (obligatorio antes de autorizar)
 * Marcar Todo: para seleccionar todas las facturas
 * Autorizar: para autorizar las facturas seleccionadas, completando el CAE y demás datos
 * Autorizar Lote: para autorizar en un solo lote las facturas seleccionadas
 * Grabar: para almacenar los datos procesados en el archivo de facturas
 * Previsualizar: para ver por pantalla la factura seleccionadas
 * Enviar: para envia por correo electrónico las facturas seleccionadas

Para solicitar soporte comercial, escriba a pyrece@sistemasagiles.com.ar
"""


class PyRece(gui.Controller):

    def on_load(self, event):
        self.cols = []
        self.items = []
        self.paths = [entrada]
        self.token = self.sign = ""
        self.smtp = None
        self.webservice = None
        if entrada and os.path.exists(entrada):
            self.cargar()

        self.components.cboWebservice.value = DEFAULT_WEBSERVICE
        self.on_cboWebservice_click(event)

        self.tipos = {
            1:u"Factura A",
            2:u"Notas de Débito A",
            3:u"Notas de Crédito A",
            4:u"Recibos A",
            5:u"Notas de Venta al contado A",
            6:u"Facturas B",
            7:u"Notas de Débito B",
            8:u"Notas de Crédito B",
            9:u"Recibos B",
            10:u"Notas de Venta al contado B",
            19:u"Facturas de Exportación",
            20:u"Nota de Débito por Operaciones con el Exterior",
            21:u"Nota de Crédito por Operaciones con el Exterior",
            39:u"Otros comprobantes A que cumplan con la R.G. N° 3419",
            40:u"Otros comprobantes B que cumplan con la R.G. N° 3419",
            60:u"Cuenta de Venta y Líquido producto A",
            61:u"Cuenta de Venta y Líquido producto B",
            63:u"Liquidación A",
            64:u"Liquidación B",
            11:u"Factura C",
            12:u"Nota de Débito C",
            13:u"Nota de Crédito C",
            15:u"Recibo C",
            }

        self.component.bgcolor = "light gray"
        # deshabilito ordenar
        ##self.components.lvwListado.GetColumnSorter = lambda: lambda x,y: 0

    def set_cols(self, cols):
        self.__cols = cols
        lv = self.components.lvwListado
        # remove old columns:
        lv.clear_all()
        # insert new columns
        for col in cols:
            ch = gui.ListColumn(lv, name=col, text=col.replace("_"," ").title(), align="left")
    def get_cols(self):
        return self.__cols
    cols = property(get_cols, set_cols)

    def set_items(self, items):
        cols = self.cols
        self.__items = items
        def convert_str(value):
            if value is None:
                return ''
            elif isinstance(value, str):
                return unicode(value,'latin1')
            elif isinstance(value, unicode):
                return value
            else:
                return str(value)
        self.components.lvwListado.items = [[convert_str(item[col]) for col in cols] for item in items]
        wx.SafeYield()
    def get_items(self):
        return self.__items
    items = property(get_items, set_items)

    def get_selected_items(self):
        for it in self.components.lvwListado.get_selected_items():
            yield it.index, it

    def set_selected_items(self, selected):
        for it in selected:
            it.selected = True

    def set_paths(self, paths):
        self.__paths = paths
        self.components.txtArchivo.value = ', '.join([fn for fn in paths])
    def get_paths(self):
        return self.__paths
    paths = property(get_paths, set_paths)

    def log(self, msg):
        if not isinstance(msg, unicode):
            msg = unicode(msg, "latin1","ignore")
        print("LOG", msg)
        self.components.txtEstado.value = msg + u"\n" + self.components.txtEstado.value
        wx.SafeYield()
        f = None
        try:
            f = open("pyrece.log","a")
            f.write("%s: " % (datetime.now(), ))
            f.write(msg.encode("ascii", "ignore"))
            f.write("\n\r")
        except Exception as e:
            print(e)
        finally:
            if f:
                f.close()

    def progreso(self, value):
        if self.items:
            per = (value+1)/float(len(self.items))*100
            self.components.pbProgreso.value = per
            wx.SafeYield()

    def error(self, code, text):
        ex = traceback.format_exception( sys.exc_type, sys.exc_value, sys.exc_traceback)
        self.log(''.join(ex))
        gui.alert(text, 'Error %s' % code)

    def verifica_ws(self):
        if not self.ws:
            gui.alert("Debe seleccionar el webservice a utilizar!", 'Advertencia')
            raise RuntimeError()
        if not self.token or not self.sign:
            gui.alert("Debe autenticarse con AFIP!", 'Advertencia')
            raise RuntimeError()

    def on_btnMarcarTodo_click(self, event):
        for it in self.components.lvwListado.items:
            it.selected = True

    def on_menu_consultas_dummy_click(self, event):
        ##self.verifica_ws()
        try:
            if self.webservice=="wsfe":
                results = self.client.FEDummy()
                msg = "AppServ %s\nDbServer %s\nAuthServer %s" % (
                    results.appserver, results.dbserver, results.authserver)
                location = self.ws.client.location
            elif self.webservice in ("wsfev1", "wsfexv1"):
                self.ws.Dummy()
                msg = "AppServ %s\nDbServer %s\nAuthServer %s" % (
                    self.ws.AppServerStatus, self.ws.DbServerStatus, self.ws.AuthServerStatus)
                location = self.ws.client.location
            else:
                msg = "%s no soportado" % self.webservice
                location = ""
            gui.alert(msg, location)
        except Exception as e:
            self.error(u'Excepción',unicode(str(e),"latin1","ignore"))

    def on_menu_consultas_lastCBTE_click(self, event):
        ##self.verifica_ws()
        options = [v for k,v in sorted([(k,v) for k,v in self.tipos.items()])]
        result = gui.single_choice(options, "Tipo de comprobante",
                                   u"Consulta Último Nro. Comprobante",
                )
        if not result:
            return
        tipocbte = [k for k,v in self.tipos.items() if v==result][0]
        result = gui.prompt(u"Punto de venta",
            u"Consulta Último Nro. Comprobante", '2')
        if not result:
            return
        ptovta = result

        try:
            if self.webservice=="wsfe":
                ultcmp = wsfe.recuperar_last_cmp(self.client, self.token, self.sign,
                    cuit, ptovta, tipocbte)
            elif  self.webservice=="wsfev1":
                ultcmp = "%s (wsfev1)" % self.ws.CompUltimoAutorizado(tipocbte, ptovta)
            elif  self.webservice=="wsfexv1":
                ultcmp = "%s (wsfexv1)" % self.ws.GetLastCMP(tipocbte, ptovta)

            gui.alert(u"Último comprobante: %s\n"
                u"Tipo: %s (%s)\nPunto de Venta: %s" % (ultcmp, self.tipos[tipocbte],
                    tipocbte, ptovta), u'Consulta Último Nro. Comprobante')
        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii","ignore"))
        except wsfe.WSFEError as e:
            self.error(e.code, e.msg.encode("ascii","ignore"))
        except Exception as e:
            self.error(u'Excepción',unicode(str(e),"latin1","ignore"))

    def on_menu_consultas_getCAE_click(self, event):
        self.verifica_ws()
        options = [v for k,v in sorted([(k,v) for k,v in self.tipos.items()])]
        result = gui.single_choice(options, "Tipo de comprobante",
            u"Consulta Comprobante",
                )
        if not result:
            return
        tipocbte = [k for k,v in self.tipos.items() if v==result][0]
        result = gui.prompt(u"Punto de venta",
            u"Consulta Comprobante", '2')
        if not result:
            return
        ptovta = result
        result = gui.prompt(u"Nº de comprobante",
            u"Consulta Comprobante", '2')
        if not result:
            return
        nrocbte = result

        try:
            if self.webservice=="wsfe":
                cae = 'no soportado'
            elif  self.webservice=="wsfev1":
                cae = "%s (wsfev1)" % self.ws.CompConsultar(tipocbte, ptovta, nrocbte)
                self.log('CAE: %s' % self.ws.CAE)
                self.log('FechaCbte: %s' % self.ws.FechaCbte)
                self.log('PuntoVenta: %s' % self.ws.PuntoVenta)
                self.log('CbteNro: %s' % self.ws.CbteNro)
                self.log('ImpTotal: %s' % self.ws.ImpTotal)
                self.log('ImpNeto: %s' % self.ws.ImpNeto)
                self.log('ImptoLiq: %s' % self.ws.ImptoLiq)
                self.log('EmisionTipo: %s' % self.ws.EmisionTipo)
            elif  self.webservice=="wsfexv1":
                cae = "%s (wsfexv1)" % self.ws.GetCMP(tipocbte, ptovta, nrocbte)
                self.log('CAE: %s' % self.ws.CAE)
                self.log('FechaCbte: %s' % self.ws.FechaCbte)
                self.log('PuntoVenta: %s' % self.ws.PuntoVenta)
                self.log('CbteNro: %s' % self.ws.CbteNro)
                self.log('ImpTotal: %s' % self.ws.ImpTotal)

            gui.alert(u"CAE: %s\n"
                u"Tipo: %s (%s)\nPunto de Venta: %s\nNumero: %s\nFecha: %s" % (
                    cae, self.tipos[tipocbte],
                    tipocbte, ptovta, nrocbte, self.ws.FechaCbte),
                    u'Consulta Comprobante')

        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii","ignore"))
        except wsfe.WSFEError as e:
            self.error(e.code, e.msg.encode("ascii","ignore"))
        except Exception as e:
            self.error(u'Excepción',unicode(str(e),"latin1","ignore"))


    def on_menu_consultas_lastID_click(self, event):
        ##self.verifica_ws()
        try:
            if self.webservice=="wsfe":
                ultnro = wsfe.ultnro(self.client, self.token, self.sign, cuit)
                print("ultnro", ultnro)
                print(self.client.xml_response)
            elif self.webservice=="wsfexv1":
                ultnro = self.ws.GetLastID()
            else:
                ultnro = None
            gui.alert(u"Último ID (máximo): %s" % (ultnro),
                u'Consulta Último ID')
        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii","ignore"))
        except wsfe.WSFEError as e:
            self.error(e.code, e.msg.encode("ascii","ignore"))
        except Exception as e:
            self.error(u'Excepción',unicode(e))


    def on_menu_ayuda_acercade_click(self, event):
        text = ACERCA_DE
        gui.alert(text, u'Acerca de PyRece Versión %s' % __version__)

    def on_menu_ayuda_instructivo_click(self, event):
        text = INSTRUCTIVO
        gui.alert(text, u'Instructivo de PyRece')

    def on_menu_ayuda_limpiar_click(self, event):
        self.components.txtEstado.value = ""

    def on_menu_ayuda_mensajesXML_click(self, event):
        self.verifica_ws()
        self.components.txtEstado.value = u"XmlRequest:\n%s\n\nXmlResponse:\n%s" % (
            self.ws.xml_request, self.ws.xml_response)
        self.component.size = (592, 517)

    def on_menu_ayuda_estado_click(self, event):
        if self.component.size[1]<517:
            self.component.size = (592, 517)
        else:
            self.component.size = (592, 265)

    def on_menu_ayuda_configuracion_click(self, event):
        self.components.txtEstado.value = open(CONFIG_FILE).read()
        self.component.size = (592, 517)

    def on_cboWebservice_click(self, event):
        self.webservice = self.components.cboWebservice.value
        self.ws = None
        self.token = None
        self.sign = None

        if self.webservice == "wsfe":
            self.client = SoapClient(wsfe_url, action=wsfe.SOAP_ACTION, namespace=wsfe.SOAP_NS,
                        trace=False, exceptions=True)
        elif self.webservice == "wsfev1":
            self.ws = wsfev1.WSFEv1()
        elif self.webservice == "wsfexv1":
            self.ws = wsfexv1.WSFEXv1()

    def on_btnAutenticar_click(self, event):
        try:
            if self.webservice in ('wsfe', ):
                service = "wsfe"
            elif self.webservice in ('wsfev1', ):
                self.log("Conectando WSFEv1... " + wsfev1_url)
                self.ws.Conectar("",wsfev1_url, proxy_dict)
                self.ws.Cuit = cuit
                service = "wsfe"
            elif self.webservice in ('wsfex', 'wsfexv1'):
                self.log("Conectando WSFEXv1... " + wsfexv1_url)
                self.ws.Conectar("",wsfexv1_url, proxy_dict)
                self.ws.Cuit = cuit
                service = "wsfex"
            else:
                gui.alert('Debe seleccionar servicio web!', 'Advertencia')
                return

            self.log("Creando TRA %s ..." % service)
            ws = wsaa.WSAA()
            tra = ws.CreateTRA(service)
            self.log("Frimando TRA (CMS) con %s %s..." % (str(cert),str(privatekey)))
            cms = ws.SignTRA(str(tra),str(cert),str(privatekey))
            self.log("Llamando a WSAA... " + wsaa_url)
            ws.Conectar("", wsdl=wsaa_url, proxy=proxy_dict)
            self.log("Proxy: %s" % proxy_dict)
            xml = ws.LoginCMS(str(cms))
            self.log("Procesando respuesta...")
            if xml:
                self.token = ws.Token
                self.sign = ws.Sign
            if DEBUG:
                self.log("Token: %s" % self.token)
                self.log("Sign: %s" % self.sign)
            elif self.token and self.sign:
                self.log("Token: %s... OK" % self.token[:10])
                self.log("Sign: %s... OK" % self.sign[:10])
            if self.webservice in ("wsfev1", "wsfexv1"):
                self.ws.Token = self.token
                self.ws.Sign = self.sign

            if xml:
                gui.alert('Autenticado OK!', 'Advertencia')
            else:
                gui.alert(u'Respuesta: %s' % ws.XmlResponse, u'No se pudo autenticar: %s' % ws.Excepcion)
        except SoapFault as e:
            self.error(e.faultcode, e.faultstring.encode("ascii","ignore"))
        except Exception as e:
            self.error(u'Excepción',unicode(e))

    def examinar(self):
        filename = entrada
        wildcard = ["Planillas Excel (*.xlsx)|*.xlsx",
                    "Archivos CSV (*.csv)|*.csv",
                    "Archivos XML (*.xml)|*.xml",
                    "Archivos TXT (*.txt)|*.txt",
                    "Archivos DBF (*.dbf)|*.dbf",
                    "Archivos JSON (*.json)|*.json",
                   ]
        if entrada.endswith("xml"):
            wildcard.sort(reverse=True)

        result = gui.open_file('Abrir', 'datos', filename, '|'.join(wildcard))
        if not result:
            return
        self.paths = [result]

    def on_menu_archivo_abrir_click(self, event):
        self.examinar()
        self.cargar()

    def on_menu_archivo_cargar_click(self, event):
        self.cargar()

    def cargar(self):
        try:
            items = []
            for fn in self.paths:
                if fn.lower().endswith(".csv") or fn.lower().endswith(".xlsx"):
                    filas = formato_csv.leer(fn)
                    items.extend(filas)
                elif fn.lower().endswith(".xml"):
                    regs = formato_xml.leer(fn)
                    items.extend(formato_csv.aplanar(regs))
                elif fn.lower().endswith(".txt"):
                    regs = formato_txt.leer(fn)
                    items.extend(formato_csv.aplanar(regs))
                elif fn.lower().endswith(".dbf"):
                    reg = formato_dbf.leer(conf_dbf, carpeta=os.path.dirname(fn))
                    items.extend(formato_csv.aplanar(reg.values()))
                elif fn.lower().endswith(".json"):
                    regs = formato_json.leer(fn)
                    items.extend(formato_csv.aplanar(regs))
                else:
                    self.error(u'Formato de archivo desconocido: %s', unicode(fn))
            if len(items) < 2:
                gui.alert(u'El archivo no tiene datos válidos', 'Advertencia')
            # extraer los nombres de columnas (ignorar vacios de XLSX)
            cols = items and [str(it).strip() for it in items[0] if it] or []
            if DEBUG:
                print("Cols",cols)
            # armar diccionario por cada linea
            items = [dict([(col,item[i]) for i, col in enumerate(cols)])
                                for item in items[1:]]
            self.cols = cols
            self.items = items
        except Exception as e:
                self.error(u'Excepción',unicode(e))
                ##raise

    def on_menu_archivo_guardar_click(self, event):
            filename = entrada
            wildcard = ["Archivos CSV (*.csv)|*.csv", "Archivos XML (*.xml)|*.xml",
                        "Archivos TXT (*.txt)|*.txt", "Archivos DBF (*.dbf)|*.dbf",
                        "Archivos JSON (*.json)|*.json",
                        "Planillas Excel (*.xlsx)|*.xlsx",
                        ]
            if entrada.endswith("xml"):
                wildcard.sort(reverse=True)
            if self.paths:
                path = self.paths[0]
            else:
                path = salida
            result = gui.save_file(title='Guardar', filename=path,
                wildcard='|'.join(wildcard))
            if not result:
                return
            fn = result[0]
            self.grabar(fn)


    def grabar(self, fn=None):
        try:
            if fn is None and salida:
                if salida.startswith("-") and self.paths:
                    fn = os.path.splitext(self.paths[0])[0] + salida
                else:
                    fn = salida
            elif not fn:
                raise RuntimeError("Debe indicar un nombre de archivo para grabar")
            if fn.lower().endswith(".csv") or fn.lower().endswith(".xlsx"):
                formato_csv.escribir([self.cols] + [[item[k] for k in self.cols] for item in self.items], fn)
            else:
                regs = formato_csv.desaplanar([self.cols] + [[item[k] for k in self.cols] for item in self.items])
                if fn.endswith(".xml"):
                    formato_xml.escribir(regs, fn)
                elif fn.endswith(".txt"):
                    formato_txt.escribir(regs, fn)
                elif fn.endswith(".dbf"):
                    formato_dbf.escribir(regs, conf_dbf, carpeta=os.path.dirname(fn))
                elif fn.endswith(".json"):
                    formato_json.escribir(regs, fn)
                else:
                    self.error(u'Formato de archivo desconocido', unicode(fn))
            gui.alert(u'Se guardó con éxito el archivo:\n%s' % (unicode(fn),), 'Guardar')
        except Exception as e:
            self.error(u'Excepción',unicode(e))

    def on_btnAutorizar_click(self, event):
        self.verifica_ws()
        try:
            ok = procesadas = rechazadas = 0
            cols = self.cols
            items = []
            self.progreso(0)
            selected = []
            for i, item in self.get_selected_items():
                kargs = item.copy()
                selected.append(item)
                kargs['cbt_desde'] = kargs['cbt_hasta'] = kargs ['cbt_numero']
                for key in kargs:
                    if isinstance(kargs[key], basestring):
                        kargs[key] = kargs[key].replace(",",".")
                if self.webservice == 'wsfe':
                    if 'id' not in kargs or kargs['id'] == "":
                        id = long(kargs['cbt_desde'])
                        id += (int(kargs['tipo_cbte'])*10**4 + int(kargs['punto_vta']))*10**8
                        kargs['id'] = id
                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k,v) for k,v in kargs.items()]))
                    if not cuit in kargs:
                        kargs['cuit'] = cuit
                    ret = wsfe.aut(self.client, self.token, self.sign, **kargs)
                    kargs.update(ret)
                    del kargs['cbt_desde']
                    del kargs['cbt_hasta']
                elif self.webservice == 'wsfev1':
                    encabezado = {}
                    for k in ('concepto', 'tipo_doc', 'nro_doc', 'tipo_cbte', 'punto_vta',
                              'cbt_desde', 'cbt_hasta', 'imp_total', 'imp_tot_conc', 'imp_neto',
                              'imp_iva', 'imp_trib', 'imp_op_ex', 'fecha_cbte',
                              'moneda_id', 'moneda_ctz'):
                        encabezado[k] = kargs[k]

                    for k in ('fecha_venc_pago', 'fecha_serv_desde', 'fecha_serv_hasta'):
                        if k in kargs:
                            encabezado[k] = kargs.get(k)

                    self.ws.CrearFactura(**encabezado)

                    for l in range(1,1000):
                        k = 'tributo_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            desc = kargs[k % 'desc']
                            base_imp = kargs[k % 'base_imp']
                            alic = kargs[k % 'alic']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarTributo(id, desc, base_imp, alic, importe)
                        else:
                            break

                    for l in range(1,1000):
                        k = 'iva_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            base_imp = kargs[k % 'base_imp']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarIva(id, base_imp, importe)
                        else:
                            break

                    for l in range(1,1000):
                        k = 'cbte_asoc_%%s_%s' % l
                        if (k % 'tipo') in kargs:
                            tipo = kargs[k % 'tipo']
                            pto_vta = kargs[k % 'pto_vta']
                            nro = kargs[k % 'nro']
                            if id:
                                self.ws.AgregarCmpAsoc(tipo, pto_vta, nro)
                        else:
                            break

                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k,v) for k,v in self.ws.factura.items()]))

                    cae = self.ws.CAESolicitar()
                    kargs.update({
                        'cae': self.ws.CAE,
                        'fecha_vto': self.ws.Vencimiento,
                        'resultado': self.ws.Resultado,
                        'motivo': self.ws.Obs,
                        'reproceso': self.ws.Reproceso,
                        'err_code': self.ws.ErrCode.encode("latin1"),
                        'err_msg': self.ws.ErrMsg.encode("latin1"),
                        })
                    if self.ws.ErrMsg:
                        gui.alert(self.ws.ErrMsg, "Error AFIP")
                    if self.ws.Obs and self.ws.Obs!='00':
                        gui.alert(self.ws.Obs, u"Observación AFIP")

                elif self.webservice == 'wsfexv1':
                    kargs['cbte_nro'] = kargs ['cbt_numero']
                    kargs['permiso_existente'] = kargs['permiso_existente'] or ""
                    encabezado = {}
                    for k in ('tipo_cbte', 'punto_vta', 'cbte_nro', 'fecha_cbte',
                              'imp_total', 'tipo_expo', 'permiso_existente', 'pais_dst_cmp',
                              'nombre_cliente', 'cuit_pais_cliente', 'domicilio_cliente',
                              'id_impositivo', 'moneda_id', 'moneda_ctz',
                              'obs_comerciales', 'obs_generales', 'forma_pago', 'incoterms',
                              'idioma_cbte', 'incoterms_ds'):
                        encabezado[k] = kargs.get(k)

                    self.ws.CrearFactura(**encabezado)

                    for l in range(1,1000):
                        k = 'codigo%s' % l
                        if k in kargs:
                            codigo = kargs['codigo%s' % l]
                            ds = kargs['descripcion%s' % l]
                            qty = kargs['cantidad%s' % l]
                            umed = kargs['umed%s' % l]
                            precio = kargs['precio%s' % l]
                            importe = kargs['importe%s' % l]
                            bonif = kargs.get('bonif%s' % l)
                            self.ws.AgregarItem(codigo, ds, qty, umed, precio, importe, bonif)
                        else:
                            break

                    for l in range(1,1000):
                        k = 'cbte_asoc_%%s_%s' % l
                        if (k % 'tipo') in kargs:
                            tipo = kargs[k % 'tipo']
                            pto_vta = kargs[k % 'pto_vta']
                            nro = kargs[k % 'nro']
                            if id:
                                self.ws.AgregarCmpAsoc(tipo, pto_vta, nro)
                        else:
                            break

                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k,v) for k,v in self.ws.factura.items()]))

                    cae = self.ws.Authorize(kargs['id'])
                    kargs.update({
                        'cae': self.ws.CAE,
                        'fecha_vto': self.ws.Vencimiento,
                        'resultado': self.ws.Resultado,
                        'motivo': self.ws.Obs,
                        'reproceso': self.ws.Reproceso,
                        'err_code': self.ws.ErrCode.encode("latin1"),
                        'err_msg': self.ws.ErrMsg.encode("latin1"),
                        })
                    if self.ws.ErrMsg:
                        gui.alert(self.ws.ErrMsg, "Error AFIP")
                    if self.ws.Obs and self.ws.Obs!='00':
                        gui.alert(self.ws.Obs, u"Observación AFIP")

                # actuaizo la factura
                for k in ('cae', 'fecha_vto', 'resultado', 'motivo', 'reproceso', 'err_code', 'err_msg'):
                    if kargs.get(k):
                        item[k] = kargs[k]
                self.items[i] = item
                self.log(u"ID: %s CAE: %s Motivo: %s Reproceso: %s" % (kargs['id'], kargs['cae'], kargs['motivo'],kargs['reproceso']))
                procesadas += 1
                if kargs['resultado'] == "R":
                    rechazadas += 1
                elif kargs['resultado'] == "A":
                    ok += 1
                self.progreso(i)
            self.items = self.items
            self.set_selected_items(selected)
            self.progreso(len(self.items) - 1)
            gui.alert(u'Proceso finalizado, procesadas %d\n\n'
                    'Aceptadas: %d\n'
                    'Rechazadas: %d' % (procesadas, ok, rechazadas),
                    u'Autorización')
            self.grabar()
        except SoapFault as e:
            self.error(e.faultcode, e.faultstring.encode("ascii","ignore"))
        except wsfe.WSFEError as e:
            self.error(e.code, e.msg.encode("ascii","ignore"))
        except KeyError as e:
            self.error("Error",u'Campo obligatorio no encontrado: %s' % e)
        except Exception as e:
            self.error(u'Excepción',unicode(e))
        finally:
            if DEBUG:
                if self.webservice == 'wsfev1' and DEBUG:
                    print(self.ws.XmlRequest)
                    print(self.ws.XmlResponse)

    def on_btnAutorizarLote_click(self, event):
        self.verifica_ws()
        if not self.items: return
        try:
            #getcontext().prec = 2
            ok = 0
            rechazadas = 0
            cols = self.cols
            items = []
            self.progreso(0)
            cbt_desde = cbt_hasta = None
            datos = {
                'tipo_cbte': None,
                'punto_vta': None,
                'fecha_cbte': None,
                'fecha_venc_pago': None,
                'fecha_cbte': None,
                'fecha_venc_pago': None,
                'fecha_serv_desde': None,
                'fecha_serv_hasta': None,
                'moneda_id': None,
                'moneda_ctz': None,
                'id': None,
            }
            importes = {
                'imp_total': Decimal(0),
                'imp_tot_conc': Decimal(0),
                'imp_neto': Decimal(0),
                'imp_iva':Decimal(0),
                'imp_op_ex': Decimal(0),
                'imp_trib': Decimal(0),
            }
            for l in range(1,5):
                k = 'iva_%%s_%s' % l
                datos[k % 'id'] = None
                importes[k % 'base_imp'] = Decimal(0)
                importes[k % 'importe'] = Decimal(0)

            for l in range(1,10):
                k = 'tributo_%%s_%s' % l
                datos[k % 'id'] = None
                datos[k % 'desc'] = None
                importes[k % 'base_imp'] = Decimal(0)
                datos[k % 'alic'] = None
                importes[k % 'importe'] = Decimal(0)

            for i, item in self.get_selected_items():
                if cbt_desde is None or int(item['cbt_numero']) < cbt_desde:
                    cbt_desde = int(item['cbt_numero'])
                if cbt_hasta is None or int(item['cbt_numero']) > cbt_hasta:
                    cbt_hasta = int(item['cbt_numero'])
                for key in item:
                    if key in datos:
                        if datos[key] is None:
                            datos[key] = item[key]
                        elif datos[key] != item[key]:
                            raise RuntimeError(u"%s tiene valores distintos en el lote!" % key)
                    if key in importes and item[key]:
                        importes[key] = importes[key] + Decimal("%.2f" % float(str(item[key].replace(",","."))))

            kargs = {'cbt_desde': cbt_desde, 'cbt_hasta': cbt_hasta}
            kargs.update({'tipo_doc': 99, 'nro_doc':  '0'})
            kargs.update(datos)
            kargs.update(importes)
            if kargs['fecha_serv_desde'] and kargs['fecha_serv_hasta']:
                kargs['presta_serv'] = 1
                kargs['concepto'] = 2
            else:
                kargs['presta_serv'] = 0
                kargs['concepto'] = 1
                del kargs['fecha_serv_desde']
                del kargs['fecha_serv_hasta']

            for key, val in importes.items():
                importes[key] = val.quantize(Decimal('.01'), rounding=ROUND_DOWN)

            if 'id' not in kargs or kargs['id'] == "":
                id = long(kargs['cbt_desde'])
                id += (int(kargs['tipo_cbte'])*10**4 + int(kargs['punto_vta']))*10**8
                kargs['id'] = id

            if DEBUG:
                self.log('\n'.join(["%s='%s'" % (k,v) for k,v in kargs.items()]))
            if '--test' in sys.argv:
                kargs['cbt_desde'] = 777
                kargs['fecha_cbte'] = '20110802'
                kargs['fecha_venc_pago'] = '20110831'

            if gui.confirm("Confirma Lote:\n"
                "Tipo: %(tipo_cbte)s Desde: %(cbt_desde)s Hasta %(cbt_hasta)s\n"
                "Neto: %(imp_neto)s IVA: %(imp_iva)s Trib.: %(imp_trib)s Total: %(imp_total)s"
                % kargs, "Autorizar lote:"):

                if self.webservice == 'wsfev1':
                    encabezado = {}
                    for k in ('concepto', 'tipo_doc', 'nro_doc', 'tipo_cbte', 'punto_vta',
                              'cbt_desde', 'cbt_hasta', 'imp_total', 'imp_tot_conc', 'imp_neto',
                              'imp_iva', 'imp_trib', 'imp_op_ex', 'fecha_cbte',
                              'moneda_id', 'moneda_ctz'):
                        encabezado[k] = kargs[k]

                    for k in ('fecha_venc_pago', 'fecha_serv_desde', 'fecha_serv_hasta'):
                        if k in kargs:
                            encabezado[k] = kargs.get(k)

                    self.ws.CrearFactura(**encabezado)
                    for l in range(1,1000):
                        k = 'iva_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            base_imp = kargs[k % 'base_imp']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarIva(id, base_imp, importe)
                        else:
                            break

                    for l in range(1,1000):
                        k = 'tributo_%%s_%s' % l
                        if (k % 'id') in kargs:
                            id = kargs[k % 'id']
                            desc = kargs[k % 'desc']
                            base_imp = kargs[k % 'base_imp']
                            alic = kargs[k % 'alic']
                            importe = kargs[k % 'importe']
                            if id:
                                self.ws.AgregarTributo(id, desc, base_imp, alic, importe)
                        else:
                            break

                    if DEBUG:
                        self.log('\n'.join(["%s='%s'" % (k,v) for k,v in self.ws.factura.items()]))

                    cae = self.ws.CAESolicitar()
                    kargs.update({
                        'cae': self.ws.CAE,
                        'fecha_vto': self.ws.Vencimiento,
                        'resultado': self.ws.Resultado,
                        'motivo': self.ws.Obs,
                        'reproceso': self.ws.Reproceso,
                        'err_code': self.ws.ErrCode.encode("latin1"),
                        'err_msg': self.ws.ErrMsg.encode("latin1"),
                        })
                    if self.ws.ErrMsg:
                        gui.alert(self.ws.ErrMsg, "Error AFIP")
                    if self.ws.Obs and self.ws.Obs!='00':
                        gui.alert(self.ws.Obs, u"Observación AFIP")

                for i, item in self.get_selected_items():
                    for key in ('id', 'cae', 'fecha_vto', 'resultado', 'motivo', 'reproceso', 'err_code', 'err_msg'):
                        item[key] = kargs[key]

                self.log("ID: %s CAE: %s Motivo: %s Reproceso: %s" % (kargs['id'], kargs['cae'], kargs['motivo'],kargs['reproceso']))
                if kargs['resultado'] == "R":
                    rechazadas += 1
                elif kargs['resultado'] == "A":
                    ok += 1

                self.items = self.items # refrescar, ver de corregir
                self.progreso(len(self.items))
                gui.alert('Proceso finalizado OK!\n\nAceptadas: %d\nRechazadas: %d' % (ok, rechazadas), 'Autorización')
                self.grabar()
        except SoapFault as e:
            self.log(self.client.xml_request)
            self.log(self.client.xml_response)
            self.error(e.faultcode, e.faultstring.encode("ascii","ignore"))
        except wsfe.WSFEError as e:
            self.error(e.code, e.msg.encode("ascii","ignore"))
        except Exception as e:
            self.error(u'Excepción',unicode(e))

    def on_btnPrevisualizar_click(self, event):
        try:
            j = 0
            for i, item in self.get_selected_items():
                j += 1
                archivo = self.generar_factura(item,  mostrar=(j==1))
        except Exception as e:
            print(e)
            self.error(u'Excepción', unicode(str(e), 'latin1', 'ignore'))

    def on_btnEnviar_click(self, event):
        try:
            ok = no = 0
            self.progreso(0)
            for i, item in self.get_selected_items():
                if not item['cae'] in ("", "NULL"):
                    archivo = self.generar_factura(item)
                    if item.get('email'):
                        self.enviar_mail(item,archivo)
                        ok += 1
                    else:
                        no += 1
                        self.log("No se envia factura %s por no tener EMAIL" % item['cbt_numero'])
                else:
                    self.log("No se envia factura %s por no tener CAE" % item['cbt_numero'])
                    no += 1
                self.progreso(i)
            self.progreso(len(self.items))
            gui.alert('Proceso finalizado OK!\n\nEnviados: %d\nNo enviados: %d' % (ok, no), 'Envio de Email')
        except Exception as e:
            self.error(u'Excepción',unicode(e))

    def generar_factura(self, fila, mostrar=False):

        try:
            fepdf = FEPDF()
        except Exception as e:
            print(e)
            raise
        fact = formato_csv.desaplanar([self.cols] + [[item[k] for k in self.cols] for item in [fila]])[0]
        fact['cbte_nro'] = fact['cbt_numero']
        fact['items'] = fact['detalles']

        for d in fact['datos']:
            fepdf.AgregarDato(d['campo'], d['valor'], d['pagina'])
            # por compatiblidad, completo campos anteriores
            if d['campo'] not in fact and d['valor']:
                fact[d['campo']] = d['valor']

        fepdf.factura = fact
        # cargo el formato CSV por defecto (factura.csv)
        fepdf.CargarFormato(conf_fact.get("formato", "factura.csv"))

        # establezco formatos (cantidad de decimales) según configuración:
        fepdf.FmtCantidad = conf_fact.get("fmt_cantidad", "0.2")
        fepdf.FmtPrecio = conf_fact.get("fmt_precio", "0.2")

        # datos fijos:
        fepdf.CUIT = cuit  # CUIT del emisor para código de barras
        for k, v in conf_pdf.items():
            fepdf.AgregarDato(k, v)

        fepdf.CrearPlantilla(papel=conf_fact.get("papel", "legal"),
                             orientacion=conf_fact.get("orientacion", "portrait"))
        fepdf.ProcesarPlantilla(num_copias=int(conf_fact.get("copias", 1)),
                                lineas_max=int(conf_fact.get("lineas_max", 24)),
                                qty_pos=conf_fact.get("cant_pos") or 'izq')

        salida = conf_fact.get("salida", "")
        fact = fepdf.factura
        if salida:
            pass
        elif 'pdf' in fact and fact['pdf']:
            salida = fact['pdf']
        else:
            # genero el nombre de archivo según datos de factura
            d = conf_fact.get('directorio', ".")
            clave_subdir = conf_fact.get('subdirectorio','fecha_cbte')
            if clave_subdir:
                d = os.path.join(d, item[clave_subdir])
            if not os.path.isdir(d):
                os.mkdir(d)
            fs = conf_fact.get('archivo','numero').split(",")
            it = item.copy()
            tipo_fact, letra_fact, numero_fact = fact['_fmt_fact']
            it['tipo'] = tipo_fact.replace(" ", "_")
            it['letra'] = letra_fact
            it['numero'] = numero_fact
            it['mes'] = item['fecha_cbte'][4:6]
            it['año'] = item['fecha_cbte'][0:4]
            # remover acentos, ñ del nombre de archivo (vía unicode):
            fn = u''.join([unicode(it.get(ff,ff)) for ff in fs])
            fn = unicodedata.normalize('NFKD', fn).encode('ASCII', 'ignore')
            salida = os.path.join(d, "%s.pdf" % fn)
        fepdf.GenerarPDF(archivo=salida)
        if mostrar:
            fepdf.MostrarPDF(archivo=salida,imprimir='--imprimir' in sys.argv)

        return salida

    def enviar_mail(self, item, archivo):
        archivo = self.generar_factura(item)
        if item['email']:
            msg = MIMEMultipart()
            msg['Subject'] = conf_mail['motivo'].replace("NUMERO",str(item['cbt_numero']))
            msg['From'] = conf_mail['remitente']
            msg['Reply-to'] = msg['From']
            msg['To'] = item['email']
            msg.preamble = 'Mensaje de multiples partes.\n'
            if not 'html' in conf_mail:
                part = MIMEText(conf_mail['cuerpo'])
                msg.attach(part)
            else:
                alt = MIMEMultipart('alternative')
                msg.attach(alt)
                text = MIMEText(conf_mail['cuerpo'])
                alt.attach(text)
                # We reference the image in the IMG SRC attribute by the ID we give it below
                html = MIMEText(conf_mail['html'], 'html')
                alt.attach(html)
            part = MIMEApplication(open(archivo,"rb").read())
            part.add_header('Content-Disposition', 'attachment', filename=os.path.basename(archivo))
            msg.attach(part)

            try:
                self.log("Enviando email: %s a %s" % (msg['Subject'], msg['To']))
                if not self.smtp:
                    self.smtp = SMTP(conf_mail['servidor'], conf_mail.get('puerto', 25))
                    if conf_mail['usuario'] and conf_mail['clave']:
                        self.smtp.ehlo()
                        self.smtp.login(conf_mail['usuario'], conf_mail['clave'])
                self.smtp.sendmail(msg['From'], msg['To'], msg.as_string())
            except Exception as e:
                self.error(u'Excepción',unicode(e))


if __name__ == '__main__':

    if len(sys.argv)>1 and not sys.argv[1].startswith("-"):
        CONFIG_FILE = sys.argv[1]
    config = SafeConfigParser()
    config.read(CONFIG_FILE)
    if not len(config.sections()):
        if os.path.exists(CONFIG_FILE):
            gui.alert(u"Error al cargar archivo de configuración: %s" %
                        CONFIG_FILE, "PyRece: Imposible Continuar")
        else:
            gui.alert(u"No se encuentra archivo de configuración: %s" %
                        CONFIG_FILE, "PyRece: Imposible Continuar")
        sys.exit(1)
    cert = config.get('WSAA','CERT')
    privatekey = config.get('WSAA','PRIVATEKEY')
    cuit = config.get('WSFEv1','CUIT')
    if config.has_option('WSFEv1','ENTRADA'):
        entrada = config.get('WSFEv1','ENTRADA')
    else:
        entrada = ""
    if not os.path.exists(entrada):
        entrada = "facturas.csv"
    if config.has_option('WSFEv1','SALIDA'):
        salida = config.get('WSFEv1','SALIDA')
    else:
        salida = "resultado.csv"

    if config.has_section('FACTURA'):
        conf_fact = dict(config.items('FACTURA'))
    else:
        conf_fact = {}

    conf_pdf = dict(config.items('PDF'))
    conf_mail = dict(config.items('MAIL'))

    if config.has_section('DBF'):
        conf_dbf = dict(config.items('DBF'))
    else:
        conf_dbf = {}


    if config.has_option('WSAA','URL') and not HOMO:
        wsaa_url = config.get('WSAA','URL')
    else:
        wsaa_url = wsaa.WSAAURL
    if config.has_option('WSFE','URL') and not HOMO:
        wsfe_url = config.get('WSFE','URL')
    else:
        wsfe_url = wsfe.WSFEURL

    if config.has_option('WSFEv1','URL') and not HOMO:
        wsfev1_url = config.get('WSFEv1','URL')
    else:
        wsfev1_url = wsfev1.WSDL

    if config.has_option('WSFEXv1','URL') and not HOMO:
        wsfexv1_url = config.get('WSFEXv1','URL')
    else:
        wsfexv1_url = wsfexv1.WSDL

    DEFAULT_WEBSERVICE = "wsfev1"
    if config.has_section('PYRECE'):
        DEFAULT_WEBSERVICE = config.get('PYRECE','WEBSERVICE')

    if config.has_section('PROXY'):
        proxy_dict = dict(("proxy_%s" % k,v) for k,v in config.items('PROXY'))
        proxy_dict['proxy_port'] = int(proxy_dict['proxy_port'])
    else:
        proxy_dict = {}

    c = PyRece()
    gui.main_loop()



#!/usr/bin/env python
# -*- coding: latin-1 -*-

[{'components': [{'components': [{'items': [{'label': u'Abrir',
                                             'name': 'abrir',
                                             'type': 'MenuItem'},
                                            {'label': u'ReCargar',
                                             'name': 'cargar',
                                             'type': 'MenuItem'},
                                            {'label': u'Guardar',
                                             'name': 'guardar',
                                             'type': 'MenuItem'}],
                                  'label': u'Archivo',
                                  'name': 'archivo',
                                  'type': 'Menu'},
                                 {'items': [{'label': u'Estado Servidores (Dummy)',
                                             'name': 'dummy',
                                             'type': 'MenuItem'},
                                            {'label': u'\xdalt. Cbte.',
                                             'name': 'lastCBTE',
                                             'type': 'MenuItem'},
                                            {'label': u'\xdalt. ID',
                                             'name': 'lastID',
                                             'type': 'MenuItem'},
                                            {'label': u'Recuperar CAE',
                                             'name': 'getCAE',
                                             'type': 'MenuItem'}],
                                  'label': u'Consultas',
                                  'name': 'consultas',
                                  'type': 'Menu'},
                                 {'items': [{'label': u'Instructivo',
                                             'name': 'instructivo',
                                             'type': 'MenuItem'},
                                            {'label': u'Acerca de',
                                             'name': 'acercade',
                                             'type': 'MenuItem'},
                                            {'label': u'Limpiar estado',
                                             'name': 'limpiar',
                                             'type': 'MenuItem'},
                                            {'label': u'Mensajes XML',
                                             'name': 'mensajesXML',
                                             'type': 'MenuItem'},
                                            {'label': u'Ver/Ocultar Estado',
                                             'name': 'estado',
                                             'type': 'MenuItem'},
                                            {'label': u'Ver Configuraci\xf3n',
                                             'name': 'configuracion',
                                             'type': 'MenuItem'}],
                                  'label': u'Ayuda',
                                  'name': 'ayuda',
                                  'type': 'Menu'}],
                  'name': 'menu',
                  'type': 'MenuBar'},
                 {'bgcolor': u'#F9F9F8',
                  'components': [{'left': '18',
                                  'name': 'lblWebservice',
                                  'text': u'Webservice:',
                                  'top': '10',
                                  'type': 'Label'},
                                 {'items': [u'wsfe',
                                            u'wsfev1',
                                            u'wsfexv1'],
                                  'left': '102',
                                  'name': 'cboWebservice',
                                  'selection': 0,
                                  'text': u'wsfe',
                                  'top': '4',
                                  'type': 'ComboBox',
                                  'value': u'wsfe',
                                  'width': '89'},
                                 {'label': u'Marcar Todo',
                                  'left': '313',
                                  'name': 'btnMarcarTodo',
                                  'tooltip': u'Seleccionar todas las facturas',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'label': u'Autorizar Lote',
                                  'left': '199',
                                  'name': 'btnAutorizarLote',
                                  'tooltip': u'Obtener CAE para todas las facturas',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'label': u'Previsualizar',
                                  'left': '414',
                                  'name': 'btnPrevisualizar',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'label': u'Autenticar',
                                  'left': '19',
                                  'name': 'btnAutenticar',
                                  'tooltip': u'Iniciar Sesin en la AFIP',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'font': {'size': 8, 'family': 'sans serif', 'face': u'Sans'},
                                  'height': '190',
                                  'left': '18',
                                  'multiline': True,
                                  'name': 'txtEstado',
                                  'text': u'\n',
                                  'top': '300',
                                  'type': 'TextBox',
                                  'value': u'\n',
                                  'width': '554'},
                                 {'left': '20',
                                  'name': 'lblProgreso',
                                  'text': u'Progreso:',
                                  'top': '194',
                                  'type': 'Label'},
                                 {'left': '23',
                                  'name': 'lblEstado',
                                  'text': u'Estado:',
                                  'top': '280',
                                  'type': 'Label'},
                                 {'label': u'Enviar',
                                  'left': '514',
                                  'name': 'btnEnviar',
                                  'tooltip': u'Generar y enviar mails',
                                  'top': '163',
                                  'type': 'Button',
                                  'width': '53'},
                                 {'editable': False,
                                  'left': '260',
                                  'name': 'txtArchivo',
                                  'text': u'facturas.csv',
                                  'top': '5',
                                  'type': 'TextBox',
                                  'value': u'facturas.csv',
                                  'width': '313'},
                                 {'left': '195',
                                  'name': 'lblArchivo',
                                  'text': u'Archivo:',
                                  'top': '10',
                                  'type': 'Label'},
                                 {'label': u'Autorizar',
                                  'left': '109',
                                  'name': 'btnAutorizar',
                                  'tooltip': u'Obtener CAE por cada factura',
                                  'top': '163',
                                  'type': 'Button'},
                                 {'bgcolor': u'#FFFFFF',
                                  'font': {'size': 8, 'face': u'Tahoma'},
                                  'height': '106',
                                  'item_count': 0,
                                  'left': '18',
                                  'name': 'lvwListado',
                                  'sort_column': -1,
                                  'top': '53',
                                  'type': 'ListView',
                                  'width': '556'},
                                 {'left': '18',
                                  'name': 'lblFacturas',
                                  'text': u'Facturas:',
                                  'top': '35',
                                  'type': 'Label',
                                  'width': '117'},
                                 {'bgcolor': u'#D1C2B6',
                                  'height': '16',
                                  'left': '113',
                                  'name': 'pbProgreso',
                                  'top': '195',
                                  'type': 'Gauge',
                                  'width': '453'},
                                 {'filename': u'logo-sistemasagiles.png',
                                  'height': '40',
                                  'left': '23',
                                  'name': 'image_63_220',
                                  'onmousedclick': u'import webbrowser; webbrowser.open_new("http://www.sistemasagiles.com.ar/")',
                                  'top': '224',
                                  'type': 'Image',
                                  'width': '142'},
                                 {'filename': u'logo-pyafipws.png',
                                  'left': '469',
                                  'name': 'image_432_229',
                                  'onmousedclick': u'import webbrowser; webbrowser.open_new("http://www.pyafipws.com.ar/")',
                                  'top': '212',
                                  'type': 'Image'}],
                  'fgcolor': u'#4C4C4C',
                  'height': '600',
                  'id': 196,
                  'image': '',
                  'label': u'',
                  'left': '1',
                  'name': u'panel',
                  'top': '0',
                  'type': 'Panel',
                  'width': '592'}],
  'height': '300px',
  'image': '',
  'name': 'bgTemplate',
  'title': u'Aplicativo Factura Electr\xf3nica (PyRece)',
  'type': 'Window',
  'width': '592px'}]



#!/usr/bin/python
# -*- coding: latin-1 -*-
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

"""Módulo para acceder a los datos de un contribuyente registrado en el Padrón
de AFIP (WS-SR-PADRON de AFIP). Consulta a Padrón Alcance 4 version 1.1
Consulta de Padrón Constancia Inscripción Alcance 5 version 2.0
"""

__author__ = "Mariano Reingart <reingart@gmail.com>"
__copyright__ = "Copyright (C) 2017 Mariano Reingart"
__license__ = "GPL 3.0"
__version__ = "1.03b"

import datetime
import decimal
import json
import os
import sys

from .utils import inicializar_y_capturar_excepciones, BaseWS, get_install_dir, json_serializer, abrir_conf
from configparser import SafeConfigParser
from .padron import TIPO_CLAVE, PROVINCIAS

import logging
log = logging.getLogger()

HOMO = False
LANZAR_EXCEPCIONES = True
WSDL = "https://awshomo.afip.gov.ar/sr-padron/webservices/personaServiceA4?wsdl"
CONFIG_FILE = "rece.ini"


class WSSrPadronA4(BaseWS):
    "Interfaz para el WebService de Consulta Padrón Contribuyentes Alcance 4"
    _public_methods_ = ['Consultar',
                        'AnalizarXml', 'ObtenerTagXml', 'LoadTestXML',
                        'SetParametros', 'SetTicketAcceso', 'GetParametro',
                        'Dummy', 'Conectar', 'DebugLog', 'SetTicketAcceso']
    _public_attrs_ = ['Token', 'Sign', 'Cuit',
        'AppServerStatus', 'DbServerStatus', 'AuthServerStatus',
        'XmlRequest', 'XmlResponse', 'Version', 'InstallDir',
        'LanzarExcepciones', 'Excepcion', 'Traceback',
        'Persona', 'data',
        'denominacion', 'imp_ganancias', 'imp_iva',
        'monotributo', 'integrante_soc', 'empleador',
        'actividad_monotributo', 'cat_iva', 'domicilios',
        'tipo_doc', 'nro_doc',
        'tipo_persona', 'estado', 'impuestos', 'actividades',
        'direccion', 'localidad', 'provincia', 'cod_postal',
        ]

    _reg_progid_ = "WSSrPadronA4"
    _reg_clsid_ = "{C2270008-4324-46F6-A2D3-60836EE63BD7}"

    # Variables globales para BaseWS:
    HOMO = HOMO
    WSDL = WSDL
    Version = "%s %s" % (__version__, HOMO and 'Homologación' or '')
    Reprocesar = True  # recuperar automaticamente CAE emitidos
    LanzarExcepciones = LANZAR_EXCEPCIONES
    factura = None

    def inicializar(self):
        BaseWS.inicializar(self)
        self.AppServerStatus = self.DbServerStatus = self.AuthServerStatus = None
        self.Persona = ''
        self.Reproceso = '' # no implementado
        self.cuit = self.dni = 0
        self.tipo_persona = ""                      # FISICA o JURIDICA
        self.tipo_doc = self.nro_doc = 0
        self.estado = ""                            # ACTIVO
        self.denominacion = ""
        self.direccion = self.localidad = self.provincia = self.cod_postal = ""
        self.domicilios = []
        self.impuestos = []
        self.actividades = []
        self.imp_iva = self.empleador = self.integrante_soc = self.cat_iva = ""
        self.monotributo = self.actividad_monotributo = ""
        self.data = {}
        self.errores = []

    def Dummy(self):
        "Obtener el estado de los servidores de la AFIP"
        ret = self.client.dummy()
        result = ret['dummyReturn']
        self.AppServerStatus = result['appserver']
        self.DbServerStatus = result['dbserver']
        self.AuthServerStatus = result['authserver']
        return True

    @inicializar_y_capturar_excepciones
    def Consultar(self, id_persona):
        "Devuelve el detalle de todos los datos del contribuyente solicitado"
        # llamar al webservice:
        log.warn("____ws_sr_padron CONSULTAR____")
        log.warn("___sign: " + str(self.Sign))
        log.warn("___token: " + str(self.Token))
        log.warn("___cuitRepresentada: " + str(self.Cuit))
        log.warn("___idPersona: " + str(id_persona))

        res = self.client.getPersona(
            sign=self.Sign.decode("utf-8"),
            token=self.Token.decode("utf-8"),
            cuitRepresentada=self.Cuit,
            idPersona=id_persona,
            )

        ret = res.get('personaReturn', {})
        # obtengo el resultado de AFIP (dict):
        data = ret.get('persona', None)
        if isinstance(data, list):
            data = data[0]
        self.data = data
        log.warn("Resultado del AFIP: " + str(self.data))
        # lo serializo
        self.Persona = json.dumps(self.data,
                                  default=json_serializer)
        # extraigo los campos principales:
        self.cuit = data["idPersona"]
        self.tipo_persona = data["tipoPersona"]
        self.tipo_doc = TIPO_CLAVE.get(data["tipoClave"])
        self.nro_doc = data.get("numeroDocumento")
        self.estado = data.get("estadoClave")
        if not "razonSocial" in data:
            self.denominacion = ", ".join([data.get("apellido", ""),
                                          data.get("nombre", "")])
        else:
            self.denominacion = data.get("razonSocial", "")
        # analizo el domicilio, dando prioridad al FISCAL, luego LEGAL/REAL
        domicilios = data.get("domicilio", [])
        domicilios.sort(key=lambda item: item["tipoDomicilio"] != "FISCAL")
        if domicilios:
            domicilio = domicilios[0]
            self.direccion = domicilio.get("direccion", "")
            self.localidad = domicilio.get("localidad", "")  # no usado en CABA
            self.provincia = PROVINCIAS.get(domicilio.get("idProvincia"), "")
            self.cod_postal = domicilio.get("codPostal")
        else:
            self.direccion = self.localidad = self.provincia = ""
            self.cod_postal = ""
        # retrocompatibilidad:
        self.domicilios = domicilios
        self.domicilio = "%s - %s (%s) - %s" % (
                            self.direccion, self.localidad,
                            self.cod_postal, self.provincia,)
        # analizo impuestos:
        self.impuestos = [imp["idImpuesto"] for imp in data.get("impuesto", [])
                          if imp['estado'] == 'ACTIVO']
        self.actividades = [act["idActividad"] for act in data.get("actividad", [])]
        mt = [cat for cat in data.get("categoria", [])
              if cat["idImpuesto"] in (20, 21) and cat["estado"] == "ACTIVO"]
        mt.sort(key=lambda cat: cat["idImpuesto"])
        self.analizar_datos(mt[0] if mt else {})

    def analizar_datos(self, cat_mt):
        # intenta determinar situación de IVA:
        if 32 in self.impuestos:
            self.imp_iva = "EX"
        elif 33 in self.impuestos:
            self.imp_iva = "NI"
        elif 34 in self.impuestos:
            self.imp_iva = "NA"
        else:
            self.imp_iva = "S" if 30 in self.impuestos else "N"
        self.monotributo = "S" if cat_mt else "N"
        self.actividad_monotributo = cat_mt.get("descripcionCategoria") if cat_mt else ""
        self.integrante_soc = ""
        self.empleador = "S" if 301 in self.impuestos else "N"
        # intenta determinar categoría de IVA (confirmar)
        if self.imp_iva in ('AC', 'S'):
            self.cat_iva = 1  # RI
        elif self.imp_iva == 'EX':
            self.cat_iva = 4  # EX
        elif self.monotributo == 'S':
            self.cat_iva = 6  # MT
        else:
            self.cat_iva = 5  # CF
        return True


class WSSrPadronA5(WSSrPadronA4):
    "Interfaz para el WebService de Consulta Padrón Constancia de Inscripción Alcance 5"

    _reg_progid_ = "WSSrPadronA5"
    _reg_clsid_ = "{DF7447DD-EEF3-4E6B-A93B-F969B5075EC8}"

    WSDL = WSDL.replace("personaServiceA4", "personaServiceA5")

    @inicializar_y_capturar_excepciones
    def Consultar(self, id_persona):
        "Devuelve el detalle de todos los datos del contribuyente solicitado"
        # llamar al webservice:
        res = self.client.getPersona(
            sign=self.Sign.decode("utf-8"),
            token=self.Token.decode("utf-8"),
            cuitRepresentada=self.Cuit,
            idPersona=id_persona,
            )
        ret = res.get('personaReturn', {})
        # obtengo el resultado de AFIP (dict):
        data = ret.get('datosGenerales', {})
        if isinstance(data, list):
            data = data[0]
        self.data = data
        # lo serializo
        self.Persona = json.dumps(ret,
                                  default=json_serializer)
        for er in 'errorConstancia', 'errorMonotributo', 'errorRegimenGeneral':
            if er in ret:
                self.errores.extend(ret[er])
        self.Excepcion = '\n\r'.join([er["error"] for er in self.errores])
        # extraigo los campos principales:
        self.tipo_persona = data.get("tipoPersona")
        self.tipo_doc = TIPO_CLAVE.get(data.get("tipoClave"))
        self.nro_doc = data.get("idPersona")
        self.cuit = self.nro_doc
        self.estado = data.get("estadoClave")
        self.es_sucesion = data.get("esSucesion")
        if not "razonSocial" in data:
            self.denominacion = ", ".join([data.get("apellido", ""),
                                          data.get("nombre", "")])
        else:
            self.denominacion = data.get("razonSocial", "")
        # analizo el domicilio, dando prioridad al FISCAL, luego LEGAL/REAL
        domicilio = data.get("domicilioFiscal", [])
        if domicilio:
            self.direccion = domicilio.get("direccion", "")
            self.localidad = domicilio.get("localidad", "")  # no usado en CABA
            self.provincia = PROVINCIAS.get(domicilio.get("idProvincia"), "")
            self.cod_postal = domicilio.get("codPostal")
        else:
            self.direccion = self.localidad = self.provincia = ""
            self.cod_postal = ""
        # retrocompatibilidad:
        self.domicilios = [domicilio]
        self.domicilio = "%s - %s (%s) - %s" % (
            self.direccion, self.localidad,
            self.cod_postal, self.provincia,)
        # extraer datos impositivos (inscripción / opción) para unificarlos:
        data_mt = ret.get("datosMonotributo", {})
        data_rg = ret.get("datosRegimenGeneral", {})
        # analizo impuestos:
        impuestos = data_mt.get("impuesto", []) + data_rg.get("impuesto", [])
        self.impuestos = [imp["idImpuesto"] for imp in impuestos]
        actividades = data_rg.get("actividad", []) + data_mt.get("actividadMonotributista", [])
        self.actividades = [act["idActividad"] for act in actividades]
        cat_mt = data_mt.get("categoriaMonotributo", {})
        self.analizar_datos(cat_mt)
        return not self.errores


def main():
    "Función principal de pruebas (obtener CAE)"
    import os, time
    global CONFIG_FILE

    DEBUG = '--debug' in sys.argv

    if '--constancia' in sys.argv:
        padron = WSSrPadronA5()
        SECTION = 'WS-SR-PADRON-A5'
        service = "ws_sr_constancia_inscripcion"
    else:
        padron = WSSrPadronA4()
        SECTION = 'WS-SR-PADRON-A4'
        service = "ws_sr_padron_a4"

    config = abrir_conf(CONFIG_FILE, DEBUG)
    if config.has_section('WSAA'):
        crt = config.get('WSAA', 'CERT')
        key = config.get('WSAA', 'PRIVATEKEY')
        cuit = config.get(SECTION, 'CUIT')
    else:
        crt, key = "reingart.crt", "reingart.key"
        cuit = "20267565393"
    url_wsaa = url_ws = None
    if config.has_option('WSAA','URL'):
        url_wsaa = config.get('WSAA', 'URL')
    if config.has_option(SECTION,'URL') and not HOMO:
        url_ws = config.get(SECTION, 'URL')

    # obteniendo el TA para pruebas
    from .wsaa import WSAA

    cache = ""
    ta = WSAA().Autenticar(service, crt, key, url_wsaa)

    padron.SetTicketAcceso(ta)
    padron.Cuit = cuit
    # padron.Conectar(cache, url_ws, cacert="conf/afip_ca_info.crt")
    padron.Conectar(cache, url_ws, cacert="")

    if "--dummy" in sys.argv:
        print(padron.client.help("dummy"))
        wssrpadron4.Dummy()
        print("AppServerStatus", wssrpadron4.AppServerStatus)
        print("DbServerStatus", wssrpadron4.DbServerStatus)
        print("AuthServerStatus", wssrpadron4.AuthServerStatus)

    try:

        if "--prueba" in sys.argv:
            id_persona = "20000000516"
        else:
            id_persona = len(sys.argv)>1 and sys.argv[1] or "20267565393"

        if "--testing" in sys.argv:
            padron.LoadTestXML("tests/xml/%s_resp.xml" % service)
        print("Consultando AFIP online via webservice...", end=' ')
        ok = padron.Consultar(id_persona)

        if DEBUG:
            print("Persona", padron.Persona)
            print(padron.Excepcion)

        print('ok' if ok else "error", padron.Excepcion)
        print("Denominacion:", padron.denominacion)
        print("Tipo:", padron.tipo_persona, padron.tipo_doc, padron.nro_doc)
        print("Estado:", padron.estado)
        print("Direccion:", padron.direccion)
        print("Localidad:", padron.localidad)
        print("Provincia:", padron.provincia)
        print("Codigo Postal:", padron.cod_postal)
        print("Impuestos:", padron.impuestos)
        print("Actividades:", padron.actividades)
        print("IVA", padron.imp_iva)
        print("MT", padron.monotributo, padron.actividad_monotributo)
        print("Empleador", padron.empleador)

        if padron.Excepcion:
            print("Excepcion:", padron.Excepcion)
            # ver padron.errores para el detalle

    except:
        raise
        print(padron.XmlRequest)
        print(padron.XmlResponse)


# busco el directorio de instalación (global para que no cambie si usan otra dll)
INSTALL_DIR = WSSrPadronA4.InstallDir = WSSrPadronA5.InstallDir = get_install_dir()


if __name__ == '__main__':

    if "--register" in sys.argv or "--unregister" in sys.argv:
        import win32com.server.register
        win32com.server.register.UseCommandLine(WSSrPadronA4)
        win32com.server.register.UseCommandLine(WSSrPadronA5)
    else:
        main()
